# autogenerated by __main__.py do not edit
name_map = {
    'PYQT_VERSION': 'PyQt6.QtCore',
    'PYQT_VERSION_STR': 'PyQt6.QtCore',
    'QAbstractAnimation': 'PyQt6.QtCore',
    'QAbstractButton': 'PyQt6.QtWidgets',
    'QAbstractEventDispatcher': 'PyQt6.QtCore',
    'QAbstractFileIconProvider': 'PyQt6.QtGui',
    'QAbstractGraphicsShapeItem': 'PyQt6.QtWidgets',
    'QAbstractItemDelegate': 'PyQt6.QtWidgets',
    'QAbstractItemModel': 'PyQt6.QtCore',
    'QAbstractItemView': 'PyQt6.QtWidgets',
    'QAbstractListModel': 'PyQt6.QtCore',
    'QAbstractNativeEventFilter': 'PyQt6.QtCore',
    'QAbstractNetworkCache': 'PyQt6.QtNetwork',
    'QAbstractOpenGLFunctions': 'PyQt6.QtOpenGL',
    'QAbstractPrintDialog': 'PyQt6.QtPrintSupport',
    'QAbstractProxyModel': 'PyQt6.QtCore',
    'QAbstractScrollArea': 'PyQt6.QtWidgets',
    'QAbstractSlider': 'PyQt6.QtWidgets',
    'QAbstractSocket': 'PyQt6.QtNetwork',
    'QAbstractSpinBox': 'PyQt6.QtWidgets',
    'QAbstractTableModel': 'PyQt6.QtCore',
    'QAbstractTextDocumentLayout': 'PyQt6.QtGui',
    'QAccessibilityHints': 'PyQt6.QtGui',
    'QAction': 'PyQt6.QtGui',
    'QActionEvent': 'PyQt6.QtGui',
    'QActionGroup': 'PyQt6.QtGui',
    'QAnimationGroup': 'PyQt6.QtCore',
    'QApplication': 'PyQt6.QtWidgets',
    'QAudio': 'PyQt6.QtMultimedia',
    'QAudioBuffer': 'PyQt6.QtMultimedia',
    'QAudioBufferInput': 'PyQt6.QtMultimedia',
    'QAudioBufferOutput': 'PyQt6.QtMultimedia',
    'QAudioDecoder': 'PyQt6.QtMultimedia',
    'QAudioDevice': 'PyQt6.QtMultimedia',
    'QAudioFormat': 'PyQt6.QtMultimedia',
    'QAudioInput': 'PyQt6.QtMultimedia',
    'QAudioOutput': 'PyQt6.QtMultimedia',
    'QAudioSink': 'PyQt6.QtMultimedia',
    'QAudioSource': 'PyQt6.QtMultimedia',
    'QAuthenticator': 'PyQt6.QtNetwork',
    'QBackingStore': 'PyQt6.QtGui',
    'QBasicTimer': 'PyQt6.QtCore',
    'QBitArray': 'PyQt6.QtCore',
    'QBitmap': 'PyQt6.QtGui',
    'QBluetoothPermission': 'PyQt6.QtCore',
    'QBoxLayout': 'PyQt6.QtWidgets',
    'QBrush': 'PyQt6.QtGui',
    'QBuffer': 'PyQt6.QtCore',
    'QButtonGroup': 'PyQt6.QtWidgets',
    'QByteArray': 'PyQt6.QtCore',
    'QByteArrayMatcher': 'PyQt6.QtCore',
    'QCalendar': 'PyQt6.QtCore',
    'QCalendarPermission': 'PyQt6.QtCore',
    'QCalendarWidget': 'PyQt6.QtWidgets',
    'QCamera': 'PyQt6.QtMultimedia',
    'QCameraDevice': 'PyQt6.QtMultimedia',
    'QCameraFormat': 'PyQt6.QtMultimedia',
    'QCameraPermission': 'PyQt6.QtCore',
    'QCapturableWindow': 'PyQt6.QtMultimedia',
    'QCborError': 'PyQt6.QtCore',
    'QCborKnownTags': 'PyQt6.QtCore',
    'QCborSimpleType': 'PyQt6.QtCore',
    'QCborStreamReader': 'PyQt6.QtCore',
    'QCborStreamWriter': 'PyQt6.QtCore',
    'QChar': 'PyQt6.QtCore',
    'QCheckBox': 'PyQt6.QtWidgets',
    'QChildEvent': 'PyQt6.QtCore',
    'QChildWindowEvent': 'PyQt6.QtGui',
    'QClipboard': 'PyQt6.QtGui',
    'QCloseEvent': 'PyQt6.QtGui',
    'QCollator': 'PyQt6.QtCore',
    'QCollatorSortKey': 'PyQt6.QtCore',
    'QColor': 'PyQt6.QtGui',
    'QColorConstants': 'PyQt6.QtGui',
    'QColorDialog': 'PyQt6.QtWidgets',
    'QColorSpace': 'PyQt6.QtGui',
    'QColorTransform': 'PyQt6.QtGui',
    'QColumnView': 'PyQt6.QtWidgets',
    'QComboBox': 'PyQt6.QtWidgets',
    'QCommandLineOption': 'PyQt6.QtCore',
    'QCommandLineParser': 'PyQt6.QtCore',
    'QCommandLinkButton': 'PyQt6.QtWidgets',
    'QCommonStyle': 'PyQt6.QtWidgets',
    'QCompleter': 'PyQt6.QtWidgets',
    'QConcatenateTablesProxyModel': 'PyQt6.QtCore',
    'QConicalGradient': 'PyQt6.QtGui',
    'QContactsPermission': 'PyQt6.QtCore',
    'QContextMenuEvent': 'PyQt6.QtGui',
    'QCoreApplication': 'PyQt6.QtCore',
    'QCryptographicHash': 'PyQt6.QtCore',
    'QCursor': 'PyQt6.QtGui',
    'QDataStream': 'PyQt6.QtCore',
    'QDataWidgetMapper': 'PyQt6.QtWidgets',
    'QDate': 'PyQt6.QtCore',
    'QDateEdit': 'PyQt6.QtWidgets',
    'QDateTime': 'PyQt6.QtCore',
    'QDateTimeEdit': 'PyQt6.QtWidgets',
    'QDeadlineTimer': 'PyQt6.QtCore',
    'QDesktopServices': 'PyQt6.QtGui',
    'QDial': 'PyQt6.QtWidgets',
    'QDialog': 'PyQt6.QtWidgets',
    'QDialogButtonBox': 'PyQt6.QtWidgets',
    'QDir': 'PyQt6.QtCore',
    'QDirIterator': 'PyQt6.QtCore',
    'QDnsDomainNameRecord': 'PyQt6.QtNetwork',
    'QDnsHostAddressRecord': 'PyQt6.QtNetwork',
    'QDnsLookup': 'PyQt6.QtNetwork',
    'QDnsMailExchangeRecord': 'PyQt6.QtNetwork',
    'QDnsServiceRecord': 'PyQt6.QtNetwork',
    'QDnsTextRecord': 'PyQt6.QtNetwork',
    'QDnsTlsAssociationRecord': 'PyQt6.QtNetwork',
    'QDockWidget': 'PyQt6.QtWidgets',
    'QDoubleSpinBox': 'PyQt6.QtWidgets',
    'QDoubleValidator': 'PyQt6.QtGui',
    'QDrag': 'PyQt6.QtGui',
    'QDragEnterEvent': 'PyQt6.QtGui',
    'QDragLeaveEvent': 'PyQt6.QtGui',
    'QDragMoveEvent': 'PyQt6.QtGui',
    'QDropEvent': 'PyQt6.QtGui',
    'QDynamicPropertyChangeEvent': 'PyQt6.QtCore',
    'QEasingCurve': 'PyQt6.QtCore',
    'QElapsedTimer': 'PyQt6.QtCore',
    'QEnterEvent': 'PyQt6.QtGui',
    'QErrorMessage': 'PyQt6.QtWidgets',
    'QEvent': 'PyQt6.QtCore',
    'QEventLoop': 'PyQt6.QtCore',
    'QEventLoopLocker': 'PyQt6.QtCore',
    'QEventPoint': 'PyQt6.QtGui',
    'QExposeEvent': 'PyQt6.QtGui',
    'QFile': 'PyQt6.QtCore',
    'QFileDevice': 'PyQt6.QtCore',
    'QFileDialog': 'PyQt6.QtWidgets',
    'QFileIconProvider': 'PyQt6.QtWidgets',
    'QFileInfo': 'PyQt6.QtCore',
    'QFileOpenEvent': 'PyQt6.QtGui',
    'QFileSelector': 'PyQt6.QtCore',
    'QFileSystemModel': 'PyQt6.QtGui',
    'QFileSystemWatcher': 'PyQt6.QtCore',
    'QFocusEvent': 'PyQt6.QtGui',
    'QFocusFrame': 'PyQt6.QtWidgets',
    'QFont': 'PyQt6.QtGui',
    'QFontComboBox': 'PyQt6.QtWidgets',
    'QFontDatabase': 'PyQt6.QtGui',
    'QFontDialog': 'PyQt6.QtWidgets',
    'QFontInfo': 'PyQt6.QtGui',
    'QFontMetrics': 'PyQt6.QtGui',
    'QFontMetricsF': 'PyQt6.QtGui',
    'QFontVariableAxis': 'PyQt6.QtGui',
    'QFormDataBuilder': 'PyQt6.QtNetwork',
    'QFormDataPartBuilder': 'PyQt6.QtNetwork',
    'QFormLayout': 'PyQt6.QtWidgets',
    'QFrame': 'PyQt6.QtWidgets',
    'QGenericArgument': 'PyQt6.QtCore',
    'QGenericReturnArgument': 'PyQt6.QtCore',
    'QGesture': 'PyQt6.QtWidgets',
    'QGestureEvent': 'PyQt6.QtWidgets',
    'QGestureRecognizer': 'PyQt6.QtWidgets',
    'QGlyphRun': 'PyQt6.QtGui',
    'QGradient': 'PyQt6.QtGui',
    'QGraphicsAnchor': 'PyQt6.QtWidgets',
    'QGraphicsAnchorLayout': 'PyQt6.QtWidgets',
    'QGraphicsBlurEffect': 'PyQt6.QtWidgets',
    'QGraphicsColorizeEffect': 'PyQt6.QtWidgets',
    'QGraphicsDropShadowEffect': 'PyQt6.QtWidgets',
    'QGraphicsEffect': 'PyQt6.QtWidgets',
    'QGraphicsEllipseItem': 'PyQt6.QtWidgets',
    'QGraphicsGridLayout': 'PyQt6.QtWidgets',
    'QGraphicsItem': 'PyQt6.QtWidgets',
    'QGraphicsItemGroup': 'PyQt6.QtWidgets',
    'QGraphicsLayout': 'PyQt6.QtWidgets',
    'QGraphicsLayoutItem': 'PyQt6.QtWidgets',
    'QGraphicsLineItem': 'PyQt6.QtWidgets',
    'QGraphicsLinearLayout': 'PyQt6.QtWidgets',
    'QGraphicsObject': 'PyQt6.QtWidgets',
    'QGraphicsOpacityEffect': 'PyQt6.QtWidgets',
    'QGraphicsPathItem': 'PyQt6.QtWidgets',
    'QGraphicsPixmapItem': 'PyQt6.QtWidgets',
    'QGraphicsPolygonItem': 'PyQt6.QtWidgets',
    'QGraphicsProxyWidget': 'PyQt6.QtWidgets',
    'QGraphicsRectItem': 'PyQt6.QtWidgets',
    'QGraphicsRotation': 'PyQt6.QtWidgets',
    'QGraphicsScale': 'PyQt6.QtWidgets',
    'QGraphicsScene': 'PyQt6.QtWidgets',
    'QGraphicsSceneContextMenuEvent': 'PyQt6.QtWidgets',
    'QGraphicsSceneDragDropEvent': 'PyQt6.QtWidgets',
    'QGraphicsSceneEvent': 'PyQt6.QtWidgets',
    'QGraphicsSceneHelpEvent': 'PyQt6.QtWidgets',
    'QGraphicsSceneHoverEvent': 'PyQt6.QtWidgets',
    'QGraphicsSceneMouseEvent': 'PyQt6.QtWidgets',
    'QGraphicsSceneMoveEvent': 'PyQt6.QtWidgets',
    'QGraphicsSceneResizeEvent': 'PyQt6.QtWidgets',
    'QGraphicsSceneWheelEvent': 'PyQt6.QtWidgets',
    'QGraphicsSimpleTextItem': 'PyQt6.QtWidgets',
    'QGraphicsTextItem': 'PyQt6.QtWidgets',
    'QGraphicsTransform': 'PyQt6.QtWidgets',
    'QGraphicsVideoItem': 'PyQt6.QtMultimediaWidgets',
    'QGraphicsView': 'PyQt6.QtWidgets',
    'QGraphicsWidget': 'PyQt6.QtWidgets',
    'QGridLayout': 'PyQt6.QtWidgets',
    'QGroupBox': 'PyQt6.QtWidgets',
    'QGuiApplication': 'PyQt6.QtGui',
    'QHBoxLayout': 'PyQt6.QtWidgets',
    'QHeaderView': 'PyQt6.QtWidgets',
    'QHelpEvent': 'PyQt6.QtGui',
    'QHideEvent': 'PyQt6.QtGui',
    'QHostAddress': 'PyQt6.QtNetwork',
    'QHostInfo': 'PyQt6.QtNetwork',
    'QHoverEvent': 'PyQt6.QtGui',
    'QHstsPolicy': 'PyQt6.QtNetwork',
    'QHttp1Configuration': 'PyQt6.QtNetwork',
    'QHttp2Configuration': 'PyQt6.QtNetwork',
    'QHttpHeaders': 'PyQt6.QtNetwork',
    'QHttpMultiPart': 'PyQt6.QtNetwork',
    'QHttpPart': 'PyQt6.QtNetwork',
    'QIODevice': 'PyQt6.QtCore',
    'QIODeviceBase': 'PyQt6.QtCore',
    'QIcon': 'PyQt6.QtGui',
    'QIconDragEvent': 'PyQt6.QtGui',
    'QIconEngine': 'PyQt6.QtGui',
    'QIdentityProxyModel': 'PyQt6.QtCore',
    'QImage': 'PyQt6.QtGui',
    'QImageCapture': 'PyQt6.QtMultimedia',
    'QImageIOHandler': 'PyQt6.QtGui',
    'QImageReader': 'PyQt6.QtGui',
    'QImageWriter': 'PyQt6.QtGui',
    'QInputDevice': 'PyQt6.QtGui',
    'QInputDialog': 'PyQt6.QtWidgets',
    'QInputEvent': 'PyQt6.QtGui',
    'QInputMethod': 'PyQt6.QtGui',
    'QInputMethodEvent': 'PyQt6.QtGui',
    'QInputMethodQueryEvent': 'PyQt6.QtGui',
    'QIntValidator': 'PyQt6.QtGui',
    'QItemDelegate': 'PyQt6.QtWidgets',
    'QItemEditorCreatorBase': 'PyQt6.QtWidgets',
    'QItemEditorFactory': 'PyQt6.QtWidgets',
    'QItemSelection': 'PyQt6.QtCore',
    'QItemSelectionModel': 'PyQt6.QtCore',
    'QItemSelectionRange': 'PyQt6.QtCore',
    'QJsonDocument': 'PyQt6.QtCore',
    'QJsonParseError': 'PyQt6.QtCore',
    'QJsonValue': 'PyQt6.QtCore',
    'QKeyCombination': 'PyQt6.QtCore',
    'QKeyEvent': 'PyQt6.QtGui',
    'QKeySequence': 'PyQt6.QtGui',
    'QKeySequenceEdit': 'PyQt6.QtWidgets',
    'QLCDNumber': 'PyQt6.QtWidgets',
    'QLabel': 'PyQt6.QtWidgets',
    'QLayout': 'PyQt6.QtWidgets',
    'QLayoutItem': 'PyQt6.QtWidgets',
    'QLibrary': 'PyQt6.QtCore',
    'QLibraryInfo': 'PyQt6.QtCore',
    'QLine': 'PyQt6.QtCore',
    'QLineEdit': 'PyQt6.QtWidgets',
    'QLineF': 'PyQt6.QtCore',
    'QLinearGradient': 'PyQt6.QtGui',
    'QListView': 'PyQt6.QtWidgets',
    'QListWidget': 'PyQt6.QtWidgets',
    'QListWidgetItem': 'PyQt6.QtWidgets',
    'QLocalServer': 'PyQt6.QtNetwork',
    'QLocalSocket': 'PyQt6.QtNetwork',
    'QLocale': 'PyQt6.QtCore',
    'QLocationPermission': 'PyQt6.QtCore',
    'QLockFile': 'PyQt6.QtCore',
    'QLoggingCategory': 'PyQt6.QtCore',
    'QMainWindow': 'PyQt6.QtWidgets',
    'QMargins': 'PyQt6.QtCore',
    'QMarginsF': 'PyQt6.QtCore',
    'QMatrix2x2': 'PyQt6.QtGui',
    'QMatrix2x3': 'PyQt6.QtGui',
    'QMatrix2x4': 'PyQt6.QtGui',
    'QMatrix3x2': 'PyQt6.QtGui',
    'QMatrix3x3': 'PyQt6.QtGui',
    'QMatrix3x4': 'PyQt6.QtGui',
    'QMatrix4x2': 'PyQt6.QtGui',
    'QMatrix4x3': 'PyQt6.QtGui',
    'QMatrix4x4': 'PyQt6.QtGui',
    'QMdiArea': 'PyQt6.QtWidgets',
    'QMdiSubWindow': 'PyQt6.QtWidgets',
    'QMediaCaptureSession': 'PyQt6.QtMultimedia',
    'QMediaDevices': 'PyQt6.QtMultimedia',
    'QMediaFormat': 'PyQt6.QtMultimedia',
    'QMediaMetaData': 'PyQt6.QtMultimedia',
    'QMediaPlayer': 'PyQt6.QtMultimedia',
    'QMediaRecorder': 'PyQt6.QtMultimedia',
    'QMediaTimeRange': 'PyQt6.QtMultimedia',
    'QMenu': 'PyQt6.QtWidgets',
    'QMenuBar': 'PyQt6.QtWidgets',
    'QMessageAuthenticationCode': 'PyQt6.QtCore',
    'QMessageBox': 'PyQt6.QtWidgets',
    'QMessageLogContext': 'PyQt6.QtCore',
    'QMessageLogger': 'PyQt6.QtCore',
    'QMetaClassInfo': 'PyQt6.QtCore',
    'QMetaEnum': 'PyQt6.QtCore',
    'QMetaMethod': 'PyQt6.QtCore',
    'QMetaObject': 'PyQt6.QtCore',
    'QMetaProperty': 'PyQt6.QtCore',
    'QMetaType': 'PyQt6.QtCore',
    'QMicrophonePermission': 'PyQt6.QtCore',
    'QMimeData': 'PyQt6.QtCore',
    'QMimeDatabase': 'PyQt6.QtCore',
    'QMimeType': 'PyQt6.QtCore',
    'QModelIndex': 'PyQt6.QtCore',
    'QModelRoleData': 'PyQt6.QtCore',
    'QModelRoleDataSpan': 'PyQt6.QtCore',
    'QMouseEvent': 'PyQt6.QtGui',
    'QMoveEvent': 'PyQt6.QtGui',
    'QMovie': 'PyQt6.QtGui',
    'QMutex': 'PyQt6.QtCore',
    'QMutexLocker': 'PyQt6.QtCore',
    'QNativeGestureEvent': 'PyQt6.QtGui',
    'QNativeInterface': 'PyQt6.QtCore',
    'QNativeIpcKey': 'PyQt6.QtCore',
    'QNetworkAccessManager': 'PyQt6.QtNetwork',
    'QNetworkAddressEntry': 'PyQt6.QtNetwork',
    'QNetworkCacheMetaData': 'PyQt6.QtNetwork',
    'QNetworkCookie': 'PyQt6.QtNetwork',
    'QNetworkCookieJar': 'PyQt6.QtNetwork',
    'QNetworkDatagram': 'PyQt6.QtNetwork',
    'QNetworkDiskCache': 'PyQt6.QtNetwork',
    'QNetworkInformation': 'PyQt6.QtNetwork',
    'QNetworkInterface': 'PyQt6.QtNetwork',
    'QNetworkProxy': 'PyQt6.QtNetwork',
    'QNetworkProxyFactory': 'PyQt6.QtNetwork',
    'QNetworkProxyQuery': 'PyQt6.QtNetwork',
    'QNetworkReply': 'PyQt6.QtNetwork',
    'QNetworkRequest': 'PyQt6.QtNetwork',
    'QNetworkRequestFactory': 'PyQt6.QtNetwork',
    'QObject': 'PyQt6.QtCore',
    'QObjectCleanupHandler': 'PyQt6.QtCore',
    'QOcspCertificateStatus': 'PyQt6.QtNetwork',
    'QOcspResponse': 'PyQt6.QtNetwork',
    'QOcspRevocationReason': 'PyQt6.QtNetwork',
    'QOffscreenSurface': 'PyQt6.QtGui',
    'QOpenGLBuffer': 'PyQt6.QtOpenGL',
    'QOpenGLContext': 'PyQt6.QtGui',
    'QOpenGLContextGroup': 'PyQt6.QtGui',
    'QOpenGLDebugLogger': 'PyQt6.QtOpenGL',
    'QOpenGLDebugMessage': 'PyQt6.QtOpenGL',
    'QOpenGLFramebufferObject': 'PyQt6.QtOpenGL',
    'QOpenGLFramebufferObjectFormat': 'PyQt6.QtOpenGL',
    'QOpenGLFunctions_2_0': 'PyQt6.QtOpenGL',
    'QOpenGLFunctions_2_1': 'PyQt6.QtOpenGL',
    'QOpenGLFunctions_4_1_Core': 'PyQt6.QtOpenGL',
    'QOpenGLPaintDevice': 'PyQt6.QtOpenGL',
    'QOpenGLPixelTransferOptions': 'PyQt6.QtOpenGL',
    'QOpenGLShader': 'PyQt6.QtOpenGL',
    'QOpenGLShaderProgram': 'PyQt6.QtOpenGL',
    'QOpenGLTexture': 'PyQt6.QtOpenGL',
    'QOpenGLTextureBlitter': 'PyQt6.QtOpenGL',
    'QOpenGLTimeMonitor': 'PyQt6.QtOpenGL',
    'QOpenGLTimerQuery': 'PyQt6.QtOpenGL',
    'QOpenGLVersionFunctionsFactory': 'PyQt6.QtOpenGL',
    'QOpenGLVersionProfile': 'PyQt6.QtOpenGL',
    'QOpenGLVertexArrayObject': 'PyQt6.QtOpenGL',
    'QOpenGLWidget': 'PyQt6.QtOpenGLWidgets',
    'QOpenGLWindow': 'PyQt6.QtOpenGL',
    'QOperatingSystemVersion': 'PyQt6.QtCore',
    'QOperatingSystemVersionBase': 'PyQt6.QtCore',
    'QPageLayout': 'PyQt6.QtGui',
    'QPageRanges': 'PyQt6.QtGui',
    'QPageSetupDialog': 'PyQt6.QtPrintSupport',
    'QPageSize': 'PyQt6.QtGui',
    'QPagedPaintDevice': 'PyQt6.QtGui',
    'QPaintDevice': 'PyQt6.QtGui',
    'QPaintDeviceWindow': 'PyQt6.QtGui',
    'QPaintEngine': 'PyQt6.QtGui',
    'QPaintEngineState': 'PyQt6.QtGui',
    'QPaintEvent': 'PyQt6.QtGui',
    'QPainter': 'PyQt6.QtGui',
    'QPainterPath': 'PyQt6.QtGui',
    'QPainterPathStroker': 'PyQt6.QtGui',
    'QPainterStateGuard': 'PyQt6.QtGui',
    'QPalette': 'PyQt6.QtGui',
    'QPanGesture': 'PyQt6.QtWidgets',
    'QParallelAnimationGroup': 'PyQt6.QtCore',
    'QPasswordDigestor': 'PyQt6.QtNetwork',
    'QPauseAnimation': 'PyQt6.QtCore',
    'QPdfOutputIntent': 'PyQt6.QtGui',
    'QPdfWriter': 'PyQt6.QtGui',
    'QPen': 'PyQt6.QtGui',
    'QPermission': 'PyQt6.QtCore',
    'QPersistentModelIndex': 'PyQt6.QtCore',
    'QPicture': 'PyQt6.QtGui',
    'QPinchGesture': 'PyQt6.QtWidgets',
    'QPixelFormat': 'PyQt6.QtGui',
    'QPixmap': 'PyQt6.QtGui',
    'QPixmapCache': 'PyQt6.QtGui',
    'QPlainTextDocumentLayout': 'PyQt6.QtWidgets',
    'QPlainTextEdit': 'PyQt6.QtWidgets',
    'QPlatformSurfaceEvent': 'PyQt6.QtGui',
    'QPlaybackOptions': 'PyQt6.QtMultimedia',
    'QPluginLoader': 'PyQt6.QtCore',
    'QPoint': 'PyQt6.QtCore',
    'QPointF': 'PyQt6.QtCore',
    'QPointerEvent': 'PyQt6.QtGui',
    'QPointingDevice': 'PyQt6.QtGui',
    'QPointingDeviceUniqueId': 'PyQt6.QtGui',
    'QPolygon': 'PyQt6.QtGui',
    'QPolygonF': 'PyQt6.QtGui',
    'QPrintDialog': 'PyQt6.QtPrintSupport',
    'QPrintEngine': 'PyQt6.QtPrintSupport',
    'QPrintPreviewDialog': 'PyQt6.QtPrintSupport',
    'QPrintPreviewWidget': 'PyQt6.QtPrintSupport',
    'QPrinter': 'PyQt6.QtPrintSupport',
    'QPrinterInfo': 'PyQt6.QtPrintSupport',
    'QProcess': 'PyQt6.QtCore',
    'QProcessEnvironment': 'PyQt6.QtCore',
    'QProgressBar': 'PyQt6.QtWidgets',
    'QProgressDialog': 'PyQt6.QtWidgets',
    'QPropertyAnimation': 'PyQt6.QtCore',
    'QProxyStyle': 'PyQt6.QtWidgets',
    'QPushButton': 'PyQt6.QtWidgets',
    'QPyAbstractRange': 'PyQt6.QtCore',
    'QPySequenceRange': 'PyQt6.QtCore',
    'QPyTableRange': 'PyQt6.QtCore',
    'QQuaternion': 'PyQt6.QtGui',
    'QQuickAsyncImageProvider': 'PyQt6.QtQuick',
    'QQuickCloseEvent': 'PyQt6.QtQuick',
    'QQuickFramebufferObject': 'PyQt6.QtQuick',
    'QQuickGraphicsConfiguration': 'PyQt6.QtQuick',
    'QQuickGraphicsDevice': 'PyQt6.QtQuick',
    'QQuickImageProvider': 'PyQt6.QtQuick',
    'QQuickImageResponse': 'PyQt6.QtQuick',
    'QQuickItem': 'PyQt6.QtQuick',
    'QQuickItemGrabResult': 'PyQt6.QtQuick',
    'QQuickPaintedItem': 'PyQt6.QtQuick',
    'QQuickRenderControl': 'PyQt6.QtQuick',
    'QQuickRenderTarget': 'PyQt6.QtQuick',
    'QQuickTextDocument': 'PyQt6.QtQuick',
    'QQuickTextureFactory': 'PyQt6.QtQuick',
    'QQuickView': 'PyQt6.QtQuick',
    'QQuickWindow': 'PyQt6.QtQuick',
    'QRadialGradient': 'PyQt6.QtGui',
    'QRadioButton': 'PyQt6.QtWidgets',
    'QRandomGenerator': 'PyQt6.QtCore',
    'QRangeModel': 'PyQt6.QtCore',
    'QRasterWindow': 'PyQt6.QtGui',
    'QRawFont': 'PyQt6.QtGui',
    'QReadLocker': 'PyQt6.QtCore',
    'QReadWriteLock': 'PyQt6.QtCore',
    'QRect': 'PyQt6.QtCore',
    'QRectF': 'PyQt6.QtCore',
    'QRecursiveMutex': 'PyQt6.QtCore',
    'QRegion': 'PyQt6.QtGui',
    'QRegularExpression': 'PyQt6.QtCore',
    'QRegularExpressionMatch': 'PyQt6.QtCore',
    'QRegularExpressionMatchIterator': 'PyQt6.QtCore',
    'QRegularExpressionValidator': 'PyQt6.QtGui',
    'QResizeEvent': 'PyQt6.QtGui',
    'QResource': 'PyQt6.QtCore',
    'QRestAccessManager': 'PyQt6.QtNetwork',
    'QRestReply': 'PyQt6.QtNetwork',
    'QRgba64': 'PyQt6.QtGui',
    'QRubberBand': 'PyQt6.QtWidgets',
    'QRunnable': 'PyQt6.QtCore',
    'QSGBasicGeometryNode': 'PyQt6.QtQuick',
    'QSGClipNode': 'PyQt6.QtQuick',
    'QSGDynamicTexture': 'PyQt6.QtQuick',
    'QSGFlatColorMaterial': 'PyQt6.QtQuick',
    'QSGGeometry': 'PyQt6.QtQuick',
    'QSGGeometryNode': 'PyQt6.QtQuick',
    'QSGImageNode': 'PyQt6.QtQuick',
    'QSGMaterial': 'PyQt6.QtQuick',
    'QSGMaterialShader': 'PyQt6.QtQuick',
    'QSGMaterialType': 'PyQt6.QtQuick',
    'QSGNode': 'PyQt6.QtQuick',
    'QSGOpacityNode': 'PyQt6.QtQuick',
    'QSGOpaqueTextureMaterial': 'PyQt6.QtQuick',
    'QSGRectangleNode': 'PyQt6.QtQuick',
    'QSGRenderNode': 'PyQt6.QtQuick',
    'QSGRendererInterface': 'PyQt6.QtQuick',
    'QSGSimpleRectNode': 'PyQt6.QtQuick',
    'QSGSimpleTextureNode': 'PyQt6.QtQuick',
    'QSGTextNode': 'PyQt6.QtQuick',
    'QSGTexture': 'PyQt6.QtQuick',
    'QSGTextureMaterial': 'PyQt6.QtQuick',
    'QSGTextureProvider': 'PyQt6.QtQuick',
    'QSGTransformNode': 'PyQt6.QtQuick',
    'QSGVertexColorMaterial': 'PyQt6.QtQuick',
    'QSaveFile': 'PyQt6.QtCore',
    'QScreen': 'PyQt6.QtGui',
    'QScreenCapture': 'PyQt6.QtMultimedia',
    'QScrollArea': 'PyQt6.QtWidgets',
    'QScrollBar': 'PyQt6.QtWidgets',
    'QScrollEvent': 'PyQt6.QtGui',
    'QScrollPrepareEvent': 'PyQt6.QtGui',
    'QScroller': 'PyQt6.QtWidgets',
    'QScrollerProperties': 'PyQt6.QtWidgets',
    'QSemaphore': 'PyQt6.QtCore',
    'QSemaphoreReleaser': 'PyQt6.QtCore',
    'QSequentialAnimationGroup': 'PyQt6.QtCore',
    'QSessionManager': 'PyQt6.QtGui',
    'QSettings': 'PyQt6.QtCore',
    'QSharedMemory': 'PyQt6.QtCore',
    'QShortcut': 'PyQt6.QtGui',
    'QShortcutEvent': 'PyQt6.QtGui',
    'QShowEvent': 'PyQt6.QtGui',
    'QSignalBlocker': 'PyQt6.QtCore',
    'QSignalMapper': 'PyQt6.QtCore',
    'QSinglePointEvent': 'PyQt6.QtGui',
    'QSize': 'PyQt6.QtCore',
    'QSizeF': 'PyQt6.QtCore',
    'QSizeGrip': 'PyQt6.QtWidgets',
    'QSizePolicy': 'PyQt6.QtWidgets',
    'QSlider': 'PyQt6.QtWidgets',
    'QSocketNotifier': 'PyQt6.QtCore',
    'QSortFilterProxyModel': 'PyQt6.QtCore',
    'QSoundEffect': 'PyQt6.QtMultimedia',
    'QSpacerItem': 'PyQt6.QtWidgets',
    'QSpinBox': 'PyQt6.QtWidgets',
    'QSplashScreen': 'PyQt6.QtWidgets',
    'QSplitter': 'PyQt6.QtWidgets',
    'QSplitterHandle': 'PyQt6.QtWidgets',
    'QSsl': 'PyQt6.QtNetwork',
    'QSslCertificate': 'PyQt6.QtNetwork',
    'QSslCertificateExtension': 'PyQt6.QtNetwork',
    'QSslCipher': 'PyQt6.QtNetwork',
    'QSslConfiguration': 'PyQt6.QtNetwork',
    'QSslDiffieHellmanParameters': 'PyQt6.QtNetwork',
    'QSslEllipticCurve': 'PyQt6.QtNetwork',
    'QSslError': 'PyQt6.QtNetwork',
    'QSslKey': 'PyQt6.QtNetwork',
    'QSslPreSharedKeyAuthenticator': 'PyQt6.QtNetwork',
    'QSslServer': 'PyQt6.QtNetwork',
    'QSslSocket': 'PyQt6.QtNetwork',
    'QStackedLayout': 'PyQt6.QtWidgets',
    'QStackedWidget': 'PyQt6.QtWidgets',
    'QStandardItem': 'PyQt6.QtGui',
    'QStandardItemModel': 'PyQt6.QtGui',
    'QStandardPaths': 'PyQt6.QtCore',
    'QStaticText': 'PyQt6.QtGui',
    'QStatusBar': 'PyQt6.QtWidgets',
    'QStatusTipEvent': 'PyQt6.QtGui',
    'QStorageInfo': 'PyQt6.QtCore',
    'QStringConverter': 'PyQt6.QtCore',
    'QStringConverterBase': 'PyQt6.QtCore',
    'QStringDecoder': 'PyQt6.QtCore',
    'QStringEncoder': 'PyQt6.QtCore',
    'QStringListModel': 'PyQt6.QtCore',
    'QStyle': 'PyQt6.QtWidgets',
    'QStyleFactory': 'PyQt6.QtWidgets',
    'QStyleHintReturn': 'PyQt6.QtWidgets',
    'QStyleHintReturnMask': 'PyQt6.QtWidgets',
    'QStyleHintReturnVariant': 'PyQt6.QtWidgets',
    'QStyleHints': 'PyQt6.QtGui',
    'QStyleOption': 'PyQt6.QtWidgets',
    'QStyleOptionButton': 'PyQt6.QtWidgets',
    'QStyleOptionComboBox': 'PyQt6.QtWidgets',
    'QStyleOptionComplex': 'PyQt6.QtWidgets',
    'QStyleOptionDockWidget': 'PyQt6.QtWidgets',
    'QStyleOptionFocusRect': 'PyQt6.QtWidgets',
    'QStyleOptionFrame': 'PyQt6.QtWidgets',
    'QStyleOptionGraphicsItem': 'PyQt6.QtWidgets',
    'QStyleOptionGroupBox': 'PyQt6.QtWidgets',
    'QStyleOptionHeader': 'PyQt6.QtWidgets',
    'QStyleOptionHeaderV2': 'PyQt6.QtWidgets',
    'QStyleOptionMenuItem': 'PyQt6.QtWidgets',
    'QStyleOptionMenuItemV2': 'PyQt6.QtWidgets',
    'QStyleOptionProgressBar': 'PyQt6.QtWidgets',
    'QStyleOptionRubberBand': 'PyQt6.QtWidgets',
    'QStyleOptionSizeGrip': 'PyQt6.QtWidgets',
    'QStyleOptionSlider': 'PyQt6.QtWidgets',
    'QStyleOptionSpinBox': 'PyQt6.QtWidgets',
    'QStyleOptionTab': 'PyQt6.QtWidgets',
    'QStyleOptionTabBarBase': 'PyQt6.QtWidgets',
    'QStyleOptionTabWidgetFrame': 'PyQt6.QtWidgets',
    'QStyleOptionTitleBar': 'PyQt6.QtWidgets',
    'QStyleOptionToolBar': 'PyQt6.QtWidgets',
    'QStyleOptionToolBox': 'PyQt6.QtWidgets',
    'QStyleOptionToolButton': 'PyQt6.QtWidgets',
    'QStyleOptionViewItem': 'PyQt6.QtWidgets',
    'QStylePainter': 'PyQt6.QtWidgets',
    'QStyledItemDelegate': 'PyQt6.QtWidgets',
    'QSurface': 'PyQt6.QtGui',
    'QSurfaceFormat': 'PyQt6.QtGui',
    'QSvgGenerator': 'PyQt6.QtSvg',
    'QSvgRenderer': 'PyQt6.QtSvg',
    'QSwipeGesture': 'PyQt6.QtWidgets',
    'QSyntaxHighlighter': 'PyQt6.QtGui',
    'QSysInfo': 'PyQt6.QtCore',
    'QSystemSemaphore': 'PyQt6.QtCore',
    'QSystemTrayIcon': 'PyQt6.QtWidgets',
    'QT_TRANSLATE_NOOP': 'PyQt6.QtCore',
    'QT_TR_NOOP': 'PyQt6.QtCore',
    'QT_VERSION': 'PyQt6.QtCore',
    'QT_VERSION_STR': 'PyQt6.QtCore',
    'QTabBar': 'PyQt6.QtWidgets',
    'QTabWidget': 'PyQt6.QtWidgets',
    'QTableView': 'PyQt6.QtWidgets',
    'QTableWidget': 'PyQt6.QtWidgets',
    'QTableWidgetItem': 'PyQt6.QtWidgets',
    'QTableWidgetSelectionRange': 'PyQt6.QtWidgets',
    'QTabletEvent': 'PyQt6.QtGui',
    'QTapAndHoldGesture': 'PyQt6.QtWidgets',
    'QTapGesture': 'PyQt6.QtWidgets',
    'QTcpServer': 'PyQt6.QtNetwork',
    'QTcpSocket': 'PyQt6.QtNetwork',
    'QTemporaryDir': 'PyQt6.QtCore',
    'QTemporaryFile': 'PyQt6.QtCore',
    'QTextBlock': 'PyQt6.QtGui',
    'QTextBlockFormat': 'PyQt6.QtGui',
    'QTextBlockGroup': 'PyQt6.QtGui',
    'QTextBlockUserData': 'PyQt6.QtGui',
    'QTextBoundaryFinder': 'PyQt6.QtCore',
    'QTextBrowser': 'PyQt6.QtWidgets',
    'QTextCharFormat': 'PyQt6.QtGui',
    'QTextCursor': 'PyQt6.QtGui',
    'QTextDocument': 'PyQt6.QtGui',
    'QTextDocumentFragment': 'PyQt6.QtGui',
    'QTextDocumentWriter': 'PyQt6.QtGui',
    'QTextEdit': 'PyQt6.QtWidgets',
    'QTextFormat': 'PyQt6.QtGui',
    'QTextFragment': 'PyQt6.QtGui',
    'QTextFrame': 'PyQt6.QtGui',
    'QTextFrameFormat': 'PyQt6.QtGui',
    'QTextImageFormat': 'PyQt6.QtGui',
    'QTextInlineObject': 'PyQt6.QtGui',
    'QTextItem': 'PyQt6.QtGui',
    'QTextLayout': 'PyQt6.QtGui',
    'QTextLength': 'PyQt6.QtGui',
    'QTextLine': 'PyQt6.QtGui',
    'QTextList': 'PyQt6.QtGui',
    'QTextListFormat': 'PyQt6.QtGui',
    'QTextObject': 'PyQt6.QtGui',
    'QTextObjectInterface': 'PyQt6.QtGui',
    'QTextOption': 'PyQt6.QtGui',
    'QTextStream': 'PyQt6.QtCore',
    'QTextStreamManipulator': 'PyQt6.QtCore',
    'QTextTable': 'PyQt6.QtGui',
    'QTextTableCell': 'PyQt6.QtGui',
    'QTextTableCellFormat': 'PyQt6.QtGui',
    'QTextTableFormat': 'PyQt6.QtGui',
    'QTextToSpeech': 'PyQt6.QtTextToSpeech',
    'QThread': 'PyQt6.QtCore',
    'QThreadPool': 'PyQt6.QtCore',
    'QTime': 'PyQt6.QtCore',
    'QTimeEdit': 'PyQt6.QtWidgets',
    'QTimeLine': 'PyQt6.QtCore',
    'QTimeZone': 'PyQt6.QtCore',
    'QTimer': 'PyQt6.QtCore',
    'QTimerEvent': 'PyQt6.QtCore',
    'QToolBar': 'PyQt6.QtWidgets',
    'QToolBox': 'PyQt6.QtWidgets',
    'QToolButton': 'PyQt6.QtWidgets',
    'QToolTip': 'PyQt6.QtWidgets',
    'QTouchEvent': 'PyQt6.QtGui',
    'QTransform': 'PyQt6.QtGui',
    'QTranslator': 'PyQt6.QtCore',
    'QTransposeProxyModel': 'PyQt6.QtCore',
    'QTreeView': 'PyQt6.QtWidgets',
    'QTreeWidget': 'PyQt6.QtWidgets',
    'QTreeWidgetItem': 'PyQt6.QtWidgets',
    'QTreeWidgetItemIterator': 'PyQt6.QtWidgets',
    'QTypeRevision': 'PyQt6.QtCore',
    'QUdpSocket': 'PyQt6.QtNetwork',
    'QUndoCommand': 'PyQt6.QtGui',
    'QUndoGroup': 'PyQt6.QtGui',
    'QUndoStack': 'PyQt6.QtGui',
    'QUndoView': 'PyQt6.QtWidgets',
    'QUrl': 'PyQt6.QtCore',
    'QUrlQuery': 'PyQt6.QtCore',
    'QUuid': 'PyQt6.QtCore',
    'QVBoxLayout': 'PyQt6.QtWidgets',
    'QValidator': 'PyQt6.QtGui',
    'QVariant': 'PyQt6.QtCore',
    'QVariantAnimation': 'PyQt6.QtCore',
    'QVector2D': 'PyQt6.QtGui',
    'QVector3D': 'PyQt6.QtGui',
    'QVector4D': 'PyQt6.QtGui',
    'QVersionNumber': 'PyQt6.QtCore',
    'QVideoFrame': 'PyQt6.QtMultimedia',
    'QVideoFrameFormat': 'PyQt6.QtMultimedia',
    'QVideoFrameInput': 'PyQt6.QtMultimedia',
    'QVideoSink': 'PyQt6.QtMultimedia',
    'QVideoWidget': 'PyQt6.QtMultimediaWidgets',
    'QVoice': 'PyQt6.QtTextToSpeech',
    'QWIDGETSIZE_MAX': 'PyQt6.QtWidgets',
    'QWaitCondition': 'PyQt6.QtCore',
    'QWhatsThis': 'PyQt6.QtWidgets',
    'QWhatsThisClickedEvent': 'PyQt6.QtGui',
    'QWheelEvent': 'PyQt6.QtGui',
    'QWidget': 'PyQt6.QtWidgets',
    'QWidgetAction': 'PyQt6.QtWidgets',
    'QWidgetItem': 'PyQt6.QtWidgets',
    'QWindow': 'PyQt6.QtGui',
    'QWindowCapture': 'PyQt6.QtMultimedia',
    'QWindowStateChangeEvent': 'PyQt6.QtGui',
    'QWizard': 'PyQt6.QtWidgets',
    'QWizardPage': 'PyQt6.QtWidgets',
    'QWriteLocker': 'PyQt6.QtCore',
    'QXmlStreamAttribute': 'PyQt6.QtCore',
    'QXmlStreamAttributes': 'PyQt6.QtCore',
    'QXmlStreamEntityDeclaration': 'PyQt6.QtCore',
    'QXmlStreamEntityResolver': 'PyQt6.QtCore',
    'QXmlStreamNamespaceDeclaration': 'PyQt6.QtCore',
    'QXmlStreamNotationDeclaration': 'PyQt6.QtCore',
    'QXmlStreamReader': 'PyQt6.QtCore',
    'QXmlStreamWriter': 'PyQt6.QtCore',
    'Q_ARG': 'PyQt6.QtCore',
    'Q_RETURN_ARG': 'PyQt6.QtCore',
    'Qt': 'PyQt6.QtCore',
    'QtAudio': 'PyQt6.QtMultimedia',
    'QtCore': 'PyQt6.QtCore',
    'QtGui': 'PyQt6.QtGui',
    'QtMsgType': 'PyQt6.QtCore',
    'QtMultimedia': 'PyQt6.QtMultimedia',
    'QtMultimediaWidgets': 'PyQt6.QtMultimediaWidgets',
    'QtNetwork': 'PyQt6.QtNetwork',
    'QtOpenGL': 'PyQt6.QtOpenGL',
    'QtOpenGLWidgets': 'PyQt6.QtOpenGLWidgets',
    'QtPrintSupport': 'PyQt6.QtPrintSupport',
    'QtQuick': 'PyQt6.QtQuick',
    'QtSvg': 'PyQt6.QtSvg',
    'QtTextToSpeech': 'PyQt6.QtTextToSpeech',
    'QtVideo': 'PyQt6.QtMultimedia',
    'QtWidgets': 'PyQt6.QtWidgets',
    'pyqtBoundSignal': 'PyQt6.QtCore',
    'pyqtClassInfo': 'PyQt6.QtCore',
    'pyqtEnum': 'PyQt6.QtCore',
    'pyqtPickleProtocol': 'PyQt6.QtCore',
    'pyqtProperty': 'PyQt6.QtCore',
    'pyqtRemoveInputHook': 'PyQt6.QtCore',
    'pyqtRestoreInputHook': 'PyQt6.QtCore',
    'pyqtSetPickleProtocol': 'PyQt6.QtCore',
    'pyqtSignal': 'PyQt6.QtCore',
    'pyqtSlot': 'PyQt6.QtCore',
    'qAbs': 'PyQt6.QtCore',
    'qAddPostRoutine': 'PyQt6.QtCore',
    'qAddPreRoutine': 'PyQt6.QtCore',
    'qAlpha': 'PyQt6.QtGui',
    'qBlue': 'PyQt6.QtGui',
    'qChecksum': 'PyQt6.QtCore',
    'qCompress': 'PyQt6.QtCore',
    'qCritical': 'PyQt6.QtCore',
    'qDebug': 'PyQt6.QtCore',
    'qDrawBorderPixmap': 'PyQt6.QtWidgets',
    'qDrawPlainRect': 'PyQt6.QtWidgets',
    'qDrawPlainRoundedRect': 'PyQt6.QtWidgets',
    'qDrawShadeLine': 'PyQt6.QtWidgets',
    'qDrawShadePanel': 'PyQt6.QtWidgets',
    'qDrawShadeRect': 'PyQt6.QtWidgets',
    'qDrawWinButton': 'PyQt6.QtWidgets',
    'qDrawWinPanel': 'PyQt6.QtWidgets',
    'qEnvironmentVariable': 'PyQt6.QtCore',
    'qEnvironmentVariableIntValue': 'PyQt6.QtCore',
    'qEnvironmentVariableIntegerValue': 'PyQt6.QtCore',
    'qEnvironmentVariableIsEmpty': 'PyQt6.QtCore',
    'qEnvironmentVariableIsSet': 'PyQt6.QtCore',
    'qFatal': 'PyQt6.QtCore',
    'qFloatDistance': 'PyQt6.QtCore',
    'qFormatLogMessage': 'PyQt6.QtCore',
    'qFuzzyCompare': 'PyQt6.QtCore',
    'qFuzzyIsNull': 'PyQt6.QtCore',
    'qGray': 'PyQt6.QtGui',
    'qGreen': 'PyQt6.QtGui',
    'qInf': 'PyQt6.QtCore',
    'qInfo': 'PyQt6.QtCore',
    'qInstallMessageHandler': 'PyQt6.QtCore',
    'qIsFinite': 'PyQt6.QtCore',
    'qIsInf': 'PyQt6.QtCore',
    'qIsNaN': 'PyQt6.QtCore',
    'qPixelFormatAlpha': 'PyQt6.QtGui',
    'qPixelFormatCmyk': 'PyQt6.QtGui',
    'qPixelFormatGrayscale': 'PyQt6.QtGui',
    'qPixelFormatHsl': 'PyQt6.QtGui',
    'qPixelFormatHsv': 'PyQt6.QtGui',
    'qPixelFormatRgba': 'PyQt6.QtGui',
    'qPixelFormatYuv': 'PyQt6.QtGui',
    'qPremultiply': 'PyQt6.QtGui',
    'qQNaN': 'PyQt6.QtCore',
    'qRed': 'PyQt6.QtGui',
    'qRegisterResourceData': 'PyQt6.QtCore',
    'qRemovePostRoutine': 'PyQt6.QtCore',
    'qRgb': 'PyQt6.QtGui',
    'qRgba': 'PyQt6.QtGui',
    'qRgba64': 'PyQt6.QtGui',
    'qRound': 'PyQt6.QtCore',
    'qRound64': 'PyQt6.QtCore',
    'qSNaN': 'PyQt6.QtCore',
    'qSetFieldWidth': 'PyQt6.QtCore',
    'qSetMessagePattern': 'PyQt6.QtCore',
    'qSetPadChar': 'PyQt6.QtCore',
    'qSetRealNumberPrecision': 'PyQt6.QtCore',
    'qUncompress': 'PyQt6.QtCore',
    'qUnpremultiply': 'PyQt6.QtGui',
    'qUnregisterResourceData': 'PyQt6.QtCore',
    'qVersion': 'PyQt6.QtCore',
    'qWarning': 'PyQt6.QtCore',
    'qYieldCpu': 'PyQt6.QtCore',
    'qt_set_sequence_auto_mnemonic': 'PyQt6.QtGui',
    'sip': 'PyQt6.sip',
}
module_names = frozenset((
    'QtCore',
    'QtGui',
    'QtWidgets',
    'QtNetwork',
    'QtSvg',
    'QtPrintSupport',
    'QtOpenGL',
    'QtOpenGLWidgets',
    'QtQuick',
    'QtMultimedia',
    'QtMultimediaWidgets',
    'QtTextToSpeech',
    'sip',
))



# License: GPLv3 Copyright: 2008, Kovid Goyal <kovid at kovidgoyal.net>

"""
Provides platform independent temporary files that persist even after
being closed.
"""

import os
import tempfile
from collections.abc import Buffer, Iterable, Iterator
from contextlib import contextmanager
from typing import IO

from calibre.constants import __appname__, filesystem_encoding, get_windows_temp_path, ismacos, iswindows
from calibre.utils.safe_atexit import remove_dir, remove_file_atexit, remove_folder_atexit, unlink

_base_dir = _osx_cache_dir = None
_prevent_recursion = False
cleanup = unlink  # some plugins import this function


def osx_cache_dir():
    global _osx_cache_dir
    if _osx_cache_dir:
        return _osx_cache_dir
    if _osx_cache_dir is None:
        _osx_cache_dir = False
        import ctypes

        libc = ctypes.CDLL(None)
        buf = ctypes.create_string_buffer(512)
        l = libc.confstr(65538, ctypes.byref(buf), len(buf))  # _CS_DARWIN_USER_CACHE_DIR = 65538
        if 0 < l < len(buf):
            try:
                q = buf.value.decode('utf-8').rstrip('\0')
            except ValueError:
                pass
            if q and os.path.isdir(q) and os.access(q, os.R_OK | os.W_OK | os.X_OK):
                _osx_cache_dir = q
                return q


get_default_tempdir = tempfile.gettempdir


def base_dir():
    global _base_dir, _prevent_recursion
    if _base_dir is not None and not os.path.exists(_base_dir):
        # Some people seem to think that running temp file cleaners that
        # delete the temp dirs of running programs is a good idea!
        if _prevent_recursion:
            _base_dir = get_default_tempdir()
        else:
            _base_dir = None
    if _base_dir is None:
        td = os.environ.get('CALIBRE_WORKER_TEMP_DIR', None)
        if td is not None:
            from calibre.utils.serialize import msgpack_loads
            from polyglot.binary import from_hex_bytes

            try:
                td = msgpack_loads(from_hex_bytes(td))
            except Exception:
                td = None
        if td and os.path.exists(td):
            _base_dir = td
        else:
            base = os.environ.get('CALIBRE_TEMP_DIR', None)
            if base is not None and iswindows:
                base = os.getenv('CALIBRE_TEMP_DIR')
            prefix = f'{__appname__}-'
            if base is None:
                if iswindows:
                    # On windows, if the TMP env var points to a path that
                    # cannot be encoded using the mbcs encoding, then the
                    # python 2 tempfile algorithm for getting the temporary
                    # directory breaks. So we use the win32 api to get a
                    # unicode temp path instead. See
                    # https://bugs.launchpad.net/bugs/937389
                    base = get_windows_temp_path()
                elif ismacos:
                    # Use the cache dir rather than the temp dir for temp files as Apple
                    # thinks deleting unused temp files is a good idea. See note under
                    # _CS_DARWIN_USER_TEMP_DIR here
                    # https://developer.apple.com/library/mac/documentation/Darwin/Reference/ManPages/man3/confstr.3.html
                    base = osx_cache_dir()

            _base_dir = tempfile.mkdtemp(prefix=prefix, dir=base or get_default_tempdir())
            orig = _prevent_recursion
            _prevent_recursion = True
            try:
                remove_folder_atexit(_base_dir)
            finally:
                _prevent_recursion = orig

    return _base_dir


def fix_tempfile_module():
    # We want the tempfile module to use base_dir() as its tempdir, but we dont
    # want to call base_dir() now as it will possibly create a tempdir, do that
    # only on demand.
    global get_default_tempdir
    if tempfile._gettempdir is not base_dir:  # type: ignore
        get_default_tempdir = tempfile._gettempdir  # type: ignore
        tempfile._gettempdir = base_dir  # type: ignore


def reset_base_dir():
    global _base_dir
    _base_dir = None


@contextmanager
def override_base_dir(newval: str) -> Iterator[None]:
    global _base_dir
    before, _base_dir = _base_dir, newval
    try:
        yield
    finally:
        _base_dir = before


def force_unicode(x):
    # Cannot use the implementation in calibre.__init__ as it causes a circular
    # dependency
    if isinstance(x, bytes):
        x = x.decode(filesystem_encoding)
    return x


def _make_file(suffix, prefix, base):
    suffix, prefix = map(force_unicode, (suffix, prefix))  # no2to3
    return tempfile.mkstemp(suffix, prefix, dir=base)


def _make_dir(suffix, prefix, base):
    suffix, prefix = map(force_unicode, (suffix, prefix))  # no2to3
    return tempfile.mkdtemp(suffix, prefix, base)


class PersistentTemporaryFile:
    """
    A file-like object that is a temporary file that is available even after being closed on
    all platforms. It is automatically deleted on normal program termination.
    """

    _file: IO[bytes]

    def __init__(self, suffix='', prefix='', dir=None, mode='w+b'):
        if prefix is None:
            prefix = ''
        if dir is None:
            dir = base_dir()
        fd, name = _make_file(suffix, prefix, dir)

        self._file = os.fdopen(fd, mode)
        self._name = name
        self._fd = fd
        remove_file_atexit(name)

    @property
    def name(self) -> str:
        return self._name

    @name.setter
    def name(self, val: str) -> None:
        self._name = val

    def read(self, size: int = -1, /) -> bytes:
        return self._file.read(size)

    def readline(self, size: int = -1, /) -> bytes:
        return self._file.readline(size)

    def readlines(self, hint: int = -1, /) -> list[bytes]:
        return self._file.readlines(hint)

    def write(self, data: Buffer, /) -> int:
        return self._file.write(data)

    def writelines(self, lines: Iterable[Buffer], /) -> None:
        self._file.writelines(lines)

    def seek(self, offset: int, whence: int = os.SEEK_SET, /) -> int:
        return self._file.seek(offset, whence)

    def tell(self) -> int:
        return self._file.tell()

    def truncate(self, size: int | None = None, /) -> int:
        return self._file.truncate(size)

    def flush(self) -> None:
        self._file.flush()

    def close(self) -> None:
        self._file.close()

    def fileno(self) -> int:
        return self._file.fileno()

    def isatty(self) -> bool:
        return self._file.isatty()

    def readable(self) -> bool:
        return self._file.readable()

    def writable(self) -> bool:
        return self._file.writable()

    def seekable(self) -> bool:
        return self._file.seekable()

    @property
    def closed(self) -> bool:
        return self._file.closed

    def __iter__(self) -> Iterator[bytes]:
        return iter(self._file)

    def __next__(self) -> bytes:
        return next(self._file)

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()

    def __del__(self):
        try:
            self.close()
        except Exception:
            pass


def PersistentTemporaryDirectory(suffix='', prefix='', dir=None):
    """
    Return the path to a newly created temporary directory that will
    be automatically deleted on application exit.
    """
    if dir is None:
        dir = base_dir()
    tdir = _make_dir(suffix, prefix, dir)
    remove_folder_atexit(tdir)
    return tdir


class TemporaryDirectory:
    """
    A temporary directory to be used in a with statement.
    """

    def __init__(self, suffix='', prefix='', dir=None, keep=False):
        self.suffix = suffix
        self.prefix = prefix
        if dir is None:
            dir = base_dir()
        self.dir = dir
        self.keep = keep

    def __enter__(self):
        if not hasattr(self, 'tdir'):
            self.tdir = _make_dir(self.suffix, self.prefix, self.dir)
        return self.tdir

    def __exit__(self, *args):
        if not self.keep and os.path.exists(self.tdir):
            remove_dir(self.tdir)


class TemporaryFile:
    def __init__(self, suffix='', prefix='', dir=None, mode='w+b'):
        if prefix is None:
            prefix = ''
        if suffix is None:
            suffix = ''
        if dir is None:
            dir = base_dir()
        self.prefix, self.suffix, self.dir, self.mode = prefix, suffix, dir, mode
        self._file = None

    def __enter__(self):
        fd, name = _make_file(self.suffix, self.prefix, self.dir)
        self._file = os.fdopen(fd, self.mode)
        self._name = name
        self._file.close()
        return name

    def __exit__(self, *args):
        unlink(self._name)


class SpooledTemporaryFile(tempfile.SpooledTemporaryFile):
    def __init__(self, max_size=0, suffix='', prefix='', dir=None, mode='w+b', bufsize=-1):
        if prefix is None:
            prefix = ''
        if suffix is None:
            suffix = ''
        if dir is None:
            dir = base_dir()
        self._name = None
        tempfile.SpooledTemporaryFile.__init__(self, max_size=max_size, suffix=suffix, prefix=prefix, dir=dir, mode=mode)

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, val):
        self._name = val

    # See https://bugs.python.org/issue26175
    def readable(self):
        return self._file.readable()

    def seekable(self):
        return self._file.seekable()

    def writable(self):
        return self._file.writable()


def better_mktemp(*args, **kwargs):
    fd, path = tempfile.mkstemp(*args, **kwargs)
    os.close(fd)
    return path



#!/usr/bin/env python
# License: GPLv3 Copyright: 2012, Kovid Goyal <kovid@kovidgoyal.net>


def format_bytes(byts):
    byts = bytearray(byts)
    byts = [f'{b:x}' for b in byts]
    return ' '.join(byts)



#!/usr/bin/env python
# License: GPLv3 Copyright: 2010, Kovid Goyal <kovid at kovidgoyal.net>

import os
import re
from collections import namedtuple
from collections.abc import Callable
from contextlib import suppress
from functools import lru_cache, partial
from math import ceil
from time import monotonic
from typing import TYPE_CHECKING, Any

from qt.core import (
    QAction,
    QApplication,
    QClipboard,
    QColor,
    QDialog,
    QEasingCurve,
    QFontMetrics,
    QIcon,
    QKeySequence,
    QMenu,
    QMimeData,
    QPainter,
    QPalette,
    QPen,
    QPixmap,
    QPoint,
    QPropertyAnimation,
    QRect,
    QSize,
    QSizePolicy,
    QSplitter,
    Qt,
    QTimer,
    QUrl,
    QWidget,
    pyqtProperty,
    pyqtSignal,
)

from calibre import fit_image, sanitize_file_name
from calibre.constants import DEBUG, config_dir, ismacos, iswindows
from calibre.db.constants import DATA_DIR_NAME, DATA_FILE_PATTERN, NO_SEARCH_LINK, RESOURCE_URL_SCHEME
from calibre.ebooks import BOOK_EXTENSIONS
from calibre.ebooks.metadata.book.base import Metadata, field_metadata
from calibre.ebooks.metadata.book.render import mi_to_html
from calibre.ebooks.metadata.search_internet import all_author_searches, all_book_searches, name_for, url_for_author_search, url_for_book_search
from calibre.gui2 import (
    NO_URL_FORMATTING,
    choose_save_file,
    clip_border_radius,
    config,
    default_author_link,
    gprefs,
    pixmap_to_data,
    qapplication_or_fail,
    question_dialog,
    rating_font,
    safe_open_url,
)
from calibre.gui2.dialogs.confirm_delete import confirm
from calibre.gui2.dialogs.confirm_delete import confirm as confirm_delete
from calibre.gui2.dnd import dnd_get_files, dnd_get_image, dnd_has_extension, dnd_has_image, image_extensions
from calibre.gui2.widgets2 import HTMLDisplay
from calibre.startup import connect_lambda
from calibre.utils.config import tweaks
from calibre.utils.img import blend_image, image_from_x
from calibre.utils.localization import _, is_rtl, langnames_to_langcodes
from calibre.utils.resources import get_path as P
from calibre.utils.serialize import json_loads
from polyglot.binary import from_hex_bytes

InternetSearch = namedtuple('InternetSearch', 'author where')

if TYPE_CHECKING:
    from calibre.gui2.dialogs.book_info import BookInfo as DialogBookInfo


def db_for_mi(mi):
    from calibre.gui2.ui import get_gui

    lp = getattr(mi, 'external_library_path', None)
    if lp:
        return get_gui(fail_if_absent=True).library_broker.get_library(lp), True
    return get_gui(fail_if_absent=True).current_db, False


def set_html(mi, html, text_browser):
    book_id = getattr(mi, 'id', None)
    search_paths = []
    db, _ = db_for_mi(mi)
    if db and book_id is not None:
        try:
            path = db.abspath(book_id, index_is_id=True)
        except Exception:  # deleted book
            path = ''
        if path:
            search_paths = [path]
    text_browser.setSearchPaths(search_paths)
    text_browser.setHtml(html)


_css_cache = {}


def css(reset=False) -> str:
    if reset:
        _css_cache.clear()
    if (ans := _css_cache.get('ans')) is None:
        val = P('templates/book_details.css', data=True).decode('utf-8')
        ans = re.sub(r'/\*.*?\*/', '', val, flags=re.DOTALL)
        if iswindows:
            # On Windows the default monospace font family is Courier which is ugly
            ans = 'pre { font-family: "Segoe UI Mono", "Consolas", monospace; }\n\n' + ans
        _css_cache['ans'] = ans
    return ans


def resolve_colors(css):
    app = qapplication_or_fail()
    col = app.palette().color(QPalette.ColorRole.PlaceholderText).name() if app.is_dark_theme else '#666'
    return css.replace('palette(placeholder-text)', col)


def resolved_css():
    return resolve_colors(css())


def copy_all(text_browser):
    mf = getattr(text_browser, 'details', text_browser)
    c = QApplication.clipboard()
    assert c is not None
    md = QMimeData()
    html = mf.toHtml()
    md.setHtml(html)
    from html5_parser import parse
    from lxml import etree

    root = parse(html)
    tables = tuple(root.iterdescendants('table'))
    for tag in root.iterdescendants(('table', 'tr', 'tbody')):
        tag.tag = 'div'
    parent = root
    is_vertical = getattr(text_browser, 'vertical', True)
    if not is_vertical:
        parent = tables[1]
    for tag in parent.iterdescendants('td'):
        for child in tag.iterdescendants('br'):
            child.tag = 'span'
            child.text = '\ue000'
        tt = etree.tostring(tag, method='text', encoding='unicode')
        tag.tag = 'span'
        for child in tuple(tag):
            tag.remove(child)
        tag.text = tt.strip()
    if not is_vertical:
        for tag in root.iterdescendants('td'):
            tag.tag = 'div'
    for tag in root.iterdescendants('a'):
        tag.attrib.pop('href', None)
    from calibre.utils.html2text import html2text

    simplified_html = etree.tostring(root, encoding='unicode')
    txt = html2text(simplified_html, single_line_break=True).strip()
    txt = txt.replace('\ue000', '\n\t')
    if iswindows:
        txt = os.linesep.join(txt.splitlines())
    # print(simplified_html)
    # print(txt)
    md.setText(txt)
    c.setMimeData(md)


def create_search_internet_menu(callback: Callable[[Any], None], author=None):
    m = QMenu(_('Search the internet for the author {}').format(author) if author is not None else _('Search the internet for this book'))
    ma = m.menuAction()
    assert ma is not None
    ma.setIcon(QIcon.ic('search.png'))
    items = all_book_searches() if author is None else all_author_searches()

    def _sort_key(k):
        nfk = name_for(k)
        assert nfk is not None
        return nfk.lower()

    for k in sorted(items, key=_sort_key):
        m.addAction(QIcon.ic('search.png'), name_for(k), partial(callback, InternetSearch(author, k)))
    return m


def is_category(field):
    from calibre.db.categories import find_categories
    from calibre.gui2.ui import get_gui

    gui = get_gui(fail_if_absent=True)
    fm = gui.current_db.field_metadata
    return field in {x[0] for x in find_categories(fm) if fm.is_custom_field(x[0])}


def is_boolean(field):
    from calibre.gui2.ui import get_gui

    gui = get_gui(fail_if_absent=True)
    fm = gui.current_db.field_metadata
    return fm.get(field, {}).get('datatype') == 'bool'


def escape_for_menu(x):
    return x.replace('&', '&&')


def init_manage_action(ac, field, value):
    from calibre.library.field_metadata import category_icon_map

    ic = category_icon_map.get(field) or 'blank.png'
    ac.setIcon(QIcon.ic(ic))
    ac.setText(_('Manage %s') % escape_for_menu(value))
    ac.current_fmt = field, value
    return ac


def add_notes_context_menu_actions(menu, book_info, field, value):
    from calibre.gui2.ui import get_gui

    db = get_gui(fail_if_absent=True).current_db.new_api
    if db.field_supports_notes(field):
        item_id = db.get_item_id(field, value)
        if item_id is not None:
            val = escape_for_menu(value)

            def edit_note():
                gui = get_gui(fail_if_absent=True)
                from calibre.gui2.dialogs.edit_category_notes import EditNoteDialog

                d = EditNoteDialog(field, item_id, gui.current_db.new_api, parent=book_info)
                if d.exec() == QDialog.DialogCode.Accepted:
                    gui.do_field_item_value_changed()

            ac = menu.addAction(_('Edit note for {}').format(val))
            ac.triggered.connect(edit_note)
            ac.setIcon(QIcon.ic('edit_input.png'))

            def delete_note():
                gui = get_gui(fail_if_absent=True)
                if question_dialog(
                    gui,
                    _('Are you sure?'),
                    _('Are you sure you want to delete the note for {} from this library? There is no undo.').format(val),
                    skip_dialog_name='book_details_delete_note_context_menu',
                ):
                    db = gui.current_db.new_api
                    db.set_notes_for(field, item_id, '')
                    gui.do_field_item_value_changed()

            ac = menu.addAction(_('Delete note for {}').format(escape_for_menu(value)))
            ac.triggered.connect(delete_note)
            ac.setIcon(QIcon.ic('trash.png'))


def init_find_in_tag_browser(menu, ac, field, value):
    from calibre.gui2.ui import get_gui

    hidden_cats = get_gui(fail_if_absent=True).tags_view._model.hidden_categories
    if field not in hidden_cats:
        ac.setIcon(QIcon.ic('search.png'))
        ac.setText(_('Find %s in the Tag browser') % escape_for_menu(value))
        ac.current_fmt = field, value
        menu.addAction(ac)


def download_cover(parent, book_id, current_cover_pixmap):
    from calibre.ebooks.metadata.sources.update import update_sources
    from calibre.gui2.ui import get_gui

    update_sources()
    from calibre.gui2.metadata.single_download import CoverFetch

    db = get_gui(fail_if_absent=True).current_db.new_api
    title = db.field_for('title', book_id)
    authors = db.field_for('authors', book_id)
    identifiers = db.field_for('identifiers', book_id, default_value={})
    d = CoverFetch(current_cover_pixmap, parent)
    if d.start(title, authors, identifiers) == QDialog.DialogCode.Accepted:
        return d.cover_pixmap


def get_icon_path(f, prefix):
    from calibre.library.field_metadata import category_icon_map

    custom_icons = gprefs['tags_browser_category_icons']
    ci = custom_icons.get(prefix + f, '')
    if ci:
        icon_path = os.path.join(config_dir, 'tb_icons', ci)
    elif prefix:
        icon_path = category_icon_map['gst']
    else:
        icon_path = category_icon_map.get(f, 'search.png')
    return icon_path


def init_find_in_grouped_search(menu, field, value, book_info):
    from calibre.gui2.ui import get_gui

    db = get_gui(fail_if_absent=True).current_db
    fm = db.field_metadata
    field_name = fm.get(field, {}).get('name', None)
    if field_name is None:
        # I don't think this can ever happen, but ...
        return
    gsts = db.prefs.get('grouped_search_terms', {})
    gsts_to_show = []
    for v in gsts:
        fk = fm.search_term_to_field_key(v)
        if field in fk:
            gsts_to_show.append(v)

    if gsts_to_show:
        m = QMenu((_('Search calibre for %s') + '...') % escape_for_menu(value), menu)
        m.setIcon(QIcon.ic('search.png'))
        menu.addMenu(m)
        m.addAction(
            QIcon.ic(get_icon_path(field, '')),
            _('in category %s') % escape_for_menu(field_name),
            lambda g=field: book_info.search_requested.emit('{}:"={}"'.format(g, value.replace('"', r'\"')), ''),
        )
        for gst in gsts_to_show:
            icon_path = get_icon_path(gst, '@')
            m.addAction(
                QIcon.ic(icon_path),
                _('in grouped search %s') % gst,
                lambda g=gst: book_info.search_requested.emit('{}:"={}"'.format(g, value.replace('"', r'\"')), ''),
            )
    else:
        menu.addAction(
            QIcon.ic('search.png'),
            _('Search calibre for {val} in category {name}').format(val=escape_for_menu(value), name=escape_for_menu(field_name)),
            lambda g=field: book_info.search_requested.emit('{}:"={}"'.format(g, value.replace('"', r'\"')), ''),
        )


@lru_cache(maxsize=2)
def comments_pat():
    return re.compile(r'<!--.*?-->', re.DOTALL)


def render_html(mi, vertical, widget, all_fields=False, render_data_func=None, pref_name='book_display_fields', pref_value=None):  # {{{
    db, is_external = db_for_mi(mi)
    show_links = not is_external
    func = render_data_func or partial(render_data, vertical_fields=db.prefs.get('book_details_vertical_categories') or ())
    try:
        table, comment_fields = func(
            mi,
            all_fields=all_fields,
            show_links=show_links,
            use_roman_numbers=config['use_roman_numerals_for_series_number'],
            pref_name=pref_name,
        )
    except TypeError:
        table, comment_fields = func(
            mi,
            all_fields=all_fields,
            show_links=show_links,
            use_roman_numbers=config['use_roman_numerals_for_series_number'],
        )

    def color_to_string(col):
        ans = '#000000'
        if col.isValid():
            col = col.toRgb()
            if col.isValid():
                ans = str(col.name())
        return ans

    comments = ''
    if comment_fields:
        comments = '\n'.join(f'<div>{x}</div>' for x in comment_fields)
        # Comments cause issues with rendering in QTextBrowser
        comments = comments_pat().sub('', comments)

    html = render_parts(table, comments, vertical)
    return html, table, comments


def render_parts(table, comments, vertical):
    templ = '''\
    <html>
        <head></head>
        <body class="%s">
        %%s
        </body>
    <html>
    ''' % ('vertical' if vertical else 'horizontal')
    right_pane = comments

    if vertical:
        ans = templ % (table + right_pane)
    else:
        ans = templ % (f'<table><tr><td valign="top" width="40%">{table}</td><td valign="top" width="60%">{right_pane}</td></tr></table>')
    return ans


def get_field_list(fm, use_defaults=False, pref_name='book_display_fields', mi=None):
    db, _ = db_for_mi(mi)
    if use_defaults:
        src = db.prefs.defaults
    else:
        old_val = gprefs.get(pref_name, None)
        if old_val is not None and not db.prefs.has_setting(pref_name):
            src = gprefs
        else:
            src = db.prefs
    fieldlist = list(src[pref_name])
    names = frozenset(x[0] for x in fieldlist)
    available = frozenset(fm.displayable_field_keys())
    for field in available - names:
        fieldlist.append((field, True))
    return [(f, d) for f, d in fieldlist if f in available]


def render_data(mi, use_roman_numbers=True, all_fields=False, pref_name='book_display_fields', vertical_fields=(), show_links=True):
    field_list = get_field_list(getattr(mi, 'field_metadata', field_metadata), pref_name=pref_name, mi=mi)
    field_list = [(x, all_fields or display) for x, display in field_list]
    db, _ = db_for_mi(mi)
    db = db.new_api
    field_maps = {}

    def item_id_if_has_note(field, item_val):
        if db.field_supports_notes(field):
            nmap = field_maps.get(field)
            if nmap is None:
                nmap = field_maps[field] = db.get_item_name_map(field)
            item_id = nmap.get(item_val)
            if item_id is not None:
                if db.notes_for(field, item_id):
                    return item_id

    # Compute the size of the link and note icons by scaling the current font
    # height by the option specified in Preferences / L&F / Book details.
    icon_size = int(QFontMetrics(QApplication.font()).height() * gprefs['book_details_note_link_icon_width'])
    return mi_to_html(
        mi,
        field_list=field_list,
        use_roman_numbers=use_roman_numbers,
        rtl=is_rtl(),
        rating_font=rating_font(),
        default_author_link=default_author_link(),
        comments_heading_pos=gprefs['book_details_comments_heading_pos'],
        for_qt=True,
        vertical_fields=vertical_fields,
        show_links=show_links,
        item_id_if_has_note=item_id_if_has_note,
        link_note_icon_size=icon_size,
    )


# }}}

# Context menu {{{


def add_format_entries(menu, data, book_info, copy_menu, search_menu):
    from calibre.ebooks.oeb.polish.main import SUPPORTED
    from calibre.gui2.ui import get_gui

    book_id = int(data['book_id'])
    fmt = data['fmt']
    init_find_in_tag_browser(search_menu, book_info.find_in_tag_browser_action, 'formats', fmt)
    init_find_in_grouped_search(search_menu, 'formats', fmt, book_info)
    db = get_gui(fail_if_absent=True).current_db.new_api
    ofmt = fmt.upper() if fmt.startswith('ORIGINAL_') else 'ORIGINAL_' + fmt
    nfmt = ofmt[len('ORIGINAL_') :]
    fmts = {x.upper() for x in db.formats(book_id)}
    for a, t in [
        ('remove', _('Delete the %s format')),
        ('save', _('Save the %s format to disk')),
        ('restore', _('Restore the %s format')),
        ('compare', ''),
        ('set_cover', _('Set the book cover from the %s file')),
    ]:
        if a == 'restore' and not fmt.startswith('ORIGINAL_'):
            continue
        if a == 'compare':
            if ofmt not in fmts or nfmt not in SUPPORTED:
                continue
            t = _('Compare to the %s format') % (fmt[9:] if fmt.startswith('ORIGINAL_') else ofmt)
        else:
            t = t % fmt
        ac = getattr(book_info, f'{a}_format_action')
        ac.current_fmt = (book_id, fmt)
        ac.setText(t)
        menu.addAction(ac)
    if not fmt.upper().startswith('ORIGINAL_'):
        from calibre.gui2.open_with import edit_programs, populate_menu

        m = QMenu(_('Open %s with...') % fmt.upper())

        def connect_action(ac, entry):
            connect_lambda(ac.triggered, book_info, lambda book_info: book_info.open_with(book_id, fmt, entry))

        populate_menu(m, connect_action, fmt)
        if len(m.actions()) == 0:
            menu.addAction(
                QIcon.ic('exec.png'),
                _('Open %s with...') % fmt.upper(),
                partial(book_info.choose_open_with, book_id, fmt),
            )
        else:
            m.setIcon(QIcon.ic('exec.png'))
            m.addSeparator()
            m.addAction(
                QIcon.ic('plus.png'),
                _('Add other application for %s files...') % fmt.upper(),
                partial(book_info.choose_open_with, book_id, fmt),
            )
            m.addAction(_('Edit Open with applications...'), partial(edit_programs, fmt, book_info))
            menu.addMenu(m)
            menu.ow = m
        if fmt.upper() in SUPPORTED:
            menu.addSeparator()
            menu.addAction(QIcon.ic('edit_book.png'), _('Edit %s format') % fmt.upper(), partial(book_info.edit_fmt, book_id, fmt))
    path = data['path']
    if path:
        if data.get('fname'):
            path = os.path.join(path, data['fname'] + '.' + data['fmt'].lower())
        ac = book_info.copy_link_action
        ac.current_url = path
        ac.setText(_('Path to file'))
        copy_menu.addAction(ac)
    if db.is_fts_enabled():
        menu.addSeparator()
        menu.addAction(
            _('Re-index the {} format for full text searching').format(fmt.upper()),
            partial(book_info.reindex_fmt, book_id, fmt),
        ).setIcon(QIcon.ic('fts.png'))


def add_link_submenu(menu: QMenu, link, book_info, field='', item_name=''):
    if field and item_name:
        m = menu.addMenu(QIcon.ic('external-link'), _('Associated link'))
        assert m is not None
        m.addAction(QIcon.ic('reference'), _('Open: {}').format(link), lambda: book_info.link_clicked.emit(link))
        m.addAction(
            QIcon.ic('minus'),
            _('Remove the link').format(link),
            lambda: book_info.link_removal_requested.emit(field, item_name),
        )
    else:
        menu.addAction(QIcon.ic('external-link'), _('Open associated link').format(link), lambda: book_info.link_clicked.emit(link))


def add_item_specific_entries(menu, data, book_info, copy_menu, search_menu):
    from calibre.gui2.ui import get_gui

    search_internet_added = False
    find_action = book_info.find_in_tag_browser_action
    dt = data['type']

    def add_copy_action(name):
        def _do_copy():
            cb = qapplication_or_fail().clipboard()
            assert cb is not None
            cb.setText(name)

        copy_menu.addAction(QIcon.ic('edit-copy.png'), _('The text: {}').format(name), _do_copy)

    if dt == 'format':
        add_format_entries(menu, data, book_info, copy_menu, search_menu)
        data['reindex_fmt_added'] = True
    elif dt == 'author':
        author = data['name']
        if data['url'] not in ('calibre', NO_SEARCH_LINK):
            ac = book_info.copy_link_action
            ac.current_url = data['url']
            ac.setText(_('&Author link'))
            copy_menu.addAction(ac)
        add_copy_action(author)
        init_find_in_tag_browser(search_menu, find_action, 'authors', author)
        init_find_in_grouped_search(search_menu, 'authors', author, book_info)
        menu.addAction(init_manage_action(book_info.manage_action, 'authors', author))
        add_notes_context_menu_actions(menu, book_info, 'authors', author)
        if hasattr(book_info, 'search_internet'):
            search_menu.addSeparator()
            search_menu.sim = create_search_internet_menu(book_info.search_internet.emit, author)
            for ac in search_menu.sim.actions():
                search_menu.addAction(ac)
                ac.setText(_('Search {0} for {1}').format(ac.text(), author))
            search_internet_added = True
        if hasattr(book_info, 'remove_item_action'):
            ac = book_info.remove_item_action
            book_id = get_gui(fail_if_absent=True).library_view.current_id
            ac.data = ('authors', author, book_id)
            ac.setText(_('Remove %s from this book') % escape_for_menu(author))
            menu.addAction(ac)
        # See if we need to add a click associated link menu line for the author
        link_map = get_gui(fail_if_absent=True).current_db.new_api.get_all_link_maps_for_book(data.get('book_id', -1))
        link = link_map.get('authors', {}).get(author)
        if link:
            add_link_submenu(menu, link, book_info, 'authors', author)
    elif dt in ('path', 'devpath'):
        path = data['loc']
        ac = book_info.copy_link_action
        if isinstance(path, int):
            path = get_gui(fail_if_absent=True).current_db.abspath(path, index_is_id=True)
        ac.current_url = path
        ac.setText(_('The location of the book'))
        copy_menu.addAction(ac)
    elif dt == 'data-path':
        path = data['loc']
        ac = book_info.copy_link_action
        path = get_gui(fail_if_absent=True).current_db.abspath(data['loc'], index_is_id=True)
        if path:
            path = os.path.join(path, DATA_DIR_NAME)
            ac.current_url = path
            ac.setText(_("The location of the book's data files"))
            copy_menu.addAction(ac)
    else:
        field = data.get('field')
        if field is not None:
            book_id = int(data['book_id'])
            value = remove_value = data['value']
            remove_name = ''
            if field == 'identifiers':
                ac = book_info.copy_link_action
                ac.current_url = value
                ac.setText(_('&Identifier'))
                copy_menu.addAction(ac)
                if data.get('url'):
                    book_info.copy_identifiers_url_action.current_url = data['url']
                    copy_menu.addAction(book_info.copy_identifiers_url_action)
                remove_value = data['id_type']
                init_find_in_tag_browser(search_menu, find_action, field, remove_value)
                init_find_in_grouped_search(search_menu, field, remove_value, book_info)
                menu.addAction(book_info.edit_identifiers_action)
                remove_name = data.get('name') or value
            elif field in ('tags', 'series', 'publisher') or is_category(field):
                add_copy_action(value)
                init_find_in_tag_browser(search_menu, find_action, field, value)
                init_find_in_grouped_search(search_menu, field, value, book_info)
                menu.addAction(init_manage_action(book_info.manage_action, field, value))
                add_notes_context_menu_actions(menu, book_info, field, value)
            elif field == 'languages':
                remove_value = langnames_to_langcodes((value,)).get(value, 'Unknown')
                init_find_in_tag_browser(search_menu, find_action, field, value)
                init_find_in_grouped_search(search_menu, field, value, book_info)
            else:
                v = data.get('original_value') or data.get('value')

                def _copy_v():
                    cb = qapplication_or_fail().clipboard()
                    assert cb is not None
                    cb.setText(v)

                copy_menu.addAction(QIcon.ic('edit-copy.png'), _('The text: {}').format(v), _copy_v)
            if field not in ('size', 'id', 'last_modified', 'sort', 'series_sort', 'uuid', 'author_sort', 'pages'):
                fm = get_gui(fail_if_absent=True).current_db.new_api.field_metadata.get(field) or {}
                if fm.get('datatype') != 'composite':
                    ac = book_info.remove_item_action
                    ac.data = (field, remove_value, book_id)
                    ac.setText(_('Remove %s from this book') % escape_for_menu(remove_name or data.get('original_value') or value))
                    menu.addAction(ac)
            # See if we need to add a click associated link menu line
            link_map = get_gui(fail_if_absent=True).current_db.new_api.get_all_link_maps_for_book(data.get('book_id', -1))
            link = link_map.get(field, {}).get(value)
            if link:
                add_link_submenu(menu, link, book_info, field, value)
        else:
            v = data.get('original_value') or data.get('value')
            if v:

                def _copy_v2():
                    cb = qapplication_or_fail().clipboard()
                    assert cb is not None
                    cb.setText(v)

                copy_menu.addAction(QIcon.ic('edit-copy.png'), _('The text: {}').format(v), _copy_v2)
    return search_internet_added


def create_copy_links(menu, data=None):
    from calibre.gui2.ui import get_gui

    db = get_gui(fail_if_absent=True).current_db.new_api
    library_id = getattr(db, 'server_library_id', None)
    if not library_id:
        return
    library_id = '_hex_-' + library_id.encode('utf-8').hex()
    book_id = get_gui(fail_if_absent=True).library_view.current_id

    def copy_to_clipboard_action(menu_text, value_text, before_action=None):
        def doit():
            cb = qapplication_or_fail().clipboard()
            assert cb is not None
            cb.setText(value_text)

        if before_action is not None:
            action = QWidget(menu).addAction(QIcon.ic('edit-copy.png'), menu_text, doit)
            menu.insertAction(before_action, action)
        else:
            menu.addAction(QIcon.ic('edit-copy.png'), menu_text, doit)

    all_links = []

    def link_action(text, url):
        nonlocal all_links
        all_links.append(url)
        copy_to_clipboard_action(text, url)

    sep = menu.addSeparator()  # Note: separators are really actions

    if data and data.get('type', '') == 'cc_url' and (u := data.get('url', '')):
        link_action(_('Custom web search link'), u)
    link_action(_('Link to show book in calibre'), f'calibre://show-book/{library_id}/{book_id}')
    link_action(_('Link to show book details in a popup window'), f'calibre://book-details/{library_id}/{book_id}')
    mi = db.new_api.get_proxy_metadata(book_id)
    if mi and mi.path:
        with suppress(Exception):
            data_files = db.new_api.list_extra_files(book_id, use_cache=True, pattern=DATA_FILE_PATTERN)
            if data_files:
                data_path = os.path.join(db.backend.library_path, mi.path, DATA_DIR_NAME)
                link_action(
                    _("Link to open book's data files folder"),
                    bytes(QUrl.fromLocalFile(data_path).toEncoded()).decode('utf-8'),
                )
    if data:
        if data.get('kind', '') == 'notes':
            field = data['field']
            item_id = data['item_id']
            note_data = db.notes_data_for(field, item_id)
            if note_data is not None:
                copy_to_clipboard_action(_('HTML for note'), note_data['doc'], sep)
                copy_to_clipboard_action(_('Text for note'), note_data['searchable_text'].partition('\n')[2], sep)
            if field.startswith('#'):
                field = '_' + field[1:]
            url = f'calibre://show-note/{library_id}/{field}/id_{item_id}'
            link_action(_('Link to show note in calibre'), url)
        else:
            field = data.get('field')
            if data['type'] == 'author':
                field = 'authors'
            if (field and field in ('tags', 'series', 'publisher', 'authors')) or is_category(field):
                name = data['name' if data['type'] == 'author' else 'value']
                eq = f'{field}:"={name}"'.encode().hex()
                link_action(
                    _('Link to show books matching {} in calibre').format(name),
                    f'calibre://search/{library_id}?eq={eq}',
                )

    for fmt in db.formats(book_id):
        fmt = fmt.upper()
        link_action(_('Link to view {} format of book').format(fmt.upper()), f'calibre://view-book/{library_id}/{book_id}/{fmt}')

    if all_links:
        menu.addSeparator()
        all_links.insert(0, '')
        all_links.insert(0, mi.get('title') + ' - ' + ' & '.join(mi.get('authors')))
        link_action(_('Copy all the above links'), '\n'.join(all_links))


def details_context_menu_event(view, ev, book_info, add_popup_action=False, edit_metadata=None):
    if not (url := view.anchorAt(ev.pos())) and (dpr := ceil(view.devicePixelRatio())) > 1:
        # Attempt to compensate for high density displays. When tabbing into an
        # anchor (URL) Qt picks a point on the outer right edge for the
        # position. Theory: on high-resolution displays, the outer right edge
        # can be a few real pixels to the right of the bounding box for the
        # anchor. As a result, Qt's chosen point isn't in the box. Compensate
        # for that by moving the position slightly to the left and see if an
        # anchor is found and, if so, use it. When back-tabbing into an anchor,
        # Qt uses the left hand side of the bounding box, which isn't affected
        # by high density. Note that this compensation could cause Qt to find
        # an anchor when the user clicks in narrow empty space between anchors.
        # I think this is ok because exact pixel mouse clicking is extremely
        # difficult.
        p = ev.pos()
        p += QPoint(-dpr, 0)
        url = view.anchorAt(p)
        if DEBUG:

            def pnt_to_str(p):
                return f'{p.x()}, {p.y()}'

            def rect_to_str(p):
                return f'{p.x()}, {p.y()}, {p.width()}, {p.height()}'

            print(
                f'BD ctxt menu pos. ev.pos: ({pnt_to_str(ev.pos())}), '
                f'npos: ({pnt_to_str(p)}), cursor rect: ({rect_to_str(view.cursorRect())}), '
                f'has url: {bool(url)}'
            )
    menu = QMenu(view)
    copy_menu = menu.addMenu(QIcon.ic('edit-copy.png'), _('Copy'))
    assert copy_menu is not None
    copy_menu.addAction(QIcon.ic('edit-copy.png'), _('All book details'), partial(copy_all, view))
    if view.textCursor().hasSelection():
        copy_menu.addAction(QIcon.ic('edit-copy.png'), _('Selected text'), view.copy)
    copy_menu.addSeparator()
    copy_links_added = False
    search_internet_added = False
    search_menu = QMenu(_('Search'), menu)
    search_menu.setIcon(QIcon.ic('search.png'))
    reindex_fmt_added = False
    if url:

        def get_data():
            kind, _, rest = url.partition(':')
            data = json_loads(from_hex_bytes(rest))
            data['kind'] = kind
            return data

        if url.startswith('action:'):
            data = get_data()
            search_internet_added = add_item_specific_entries(menu, data, book_info, copy_menu, search_menu)
            create_copy_links(copy_menu, data)
            copy_links_added = True
            reindex_fmt_added = 'reindex_fmt_added' in data
        elif url.startswith('notes:'):
            data = get_data()
            create_copy_links(copy_menu, data)
            copy_links_added = True
            search_internet_added = True
            add_notes_context_menu_actions(menu, view, data['field'], data['value'])
        elif not url.startswith('#'):
            ac = book_info.copy_link_action
            ac.current_url = url
            ac.setText(_('Copy link location'))
            menu.addAction(ac)
            add_link_submenu(menu, url, book_info)
    if not copy_links_added:
        create_copy_links(copy_menu)

    if not search_internet_added and hasattr(book_info, 'search_internet'):
        sim = create_search_internet_menu(book_info.search_internet.emit)
        if search_menu.isEmpty():
            search_menu = sim
        else:
            search_menu.addSeparator()
            for ac in sim.actions():
                search_menu.addAction(ac)
                ac.setText(_('Search {0} for this book').format(ac.text()))
    if not search_menu.isEmpty():
        menu.addMenu(search_menu)
    for ac in tuple(menu.actions()):
        if not ac.isEnabled():
            menu.removeAction(ac)
    menu.addSeparator()
    from calibre.gui2.ui import get_gui

    if add_popup_action:
        menu.addMenu(get_gui(fail_if_absent=True).iactions['Show Book Details'].qaction.menu())
    # We can't open edit metadata from a locked window because EM expects to
    # be editing the current book, which this book probably isn't
    elif edit_metadata is not None:
        ema = get_gui(fail_if_absent=True).iactions['Edit Metadata'].menuless_qaction
        menu.addAction(
            _('Open the Edit metadata window') + '\t' + ema.shortcut().toString(QKeySequence.SequenceFormat.NativeText),
            edit_metadata,
        )
    menu.addSeparator()
    book_id = get_gui(fail_if_absent=True).library_view.current_id
    if not reindex_fmt_added:
        ac_fts = menu.addAction(_('Re-index this book for full text searching'), partial(book_info.reindex_fmt, book_id, ''))
        assert ac_fts is not None
        ac_fts.setIcon(QIcon.ic('fts.png'))
    ac_pages = menu.addAction(_('Re-count the pages in this book'), partial(book_info.recount_pages, book_id))
    assert ac_pages is not None
    ac_pages.setIcon(QIcon.ic('bookshelf.png'))
    if len(menu.actions()) > 0:
        menu.exec(ev.globalPos())


# }}}


def create_open_cover_with_menu(self, parent_menu, text=''):
    from calibre.gui2.open_with import edit_programs, populate_menu

    m = QMenu(text or _('Open cover with...'))

    def connect_action(ac, entry):
        connect_lambda(ac.triggered, self, lambda self: self.open_with(entry))

    populate_menu(m, connect_action, 'cover_image')
    if len(m.actions()) == 0:
        parent_menu.addAction(text or _('Open cover with...'), self.choose_open_with)
    else:
        m.addSeparator()
        m.addAction(_('Add another application to open cover with...'), self.choose_open_with)
        m.addAction(_('Edit Open with applications...'), partial(edit_programs, 'cover_image', self))
        parent_menu.ocw = m
        parent_menu.addMenu(m)
    return m


class CoverView(QWidget):  # {{{
    cover_changed = pyqtSignal(object, object)
    cover_removed = pyqtSignal(object)
    open_cover_with = pyqtSignal(object, object)
    search_internet = pyqtSignal(object)

    def __init__(self, vertical, parent=None):
        QWidget.__init__(self, parent)
        self._current_pixmap_size = QSize(120, 120)
        self.change_layout(vertical)

        self.animation = QPropertyAnimation(self, b'current_pixmap_size', self)
        self.animation.setEasingCurve(QEasingCurve(QEasingCurve.Type.OutExpo))
        self.animation.setDuration(1000)
        self.animation.setStartValue(QSize(0, 0))
        self.animation.valueChanged.connect(self.value_changed)

        self.default_pixmap = qapplication_or_fail().cached_qpixmap('default_cover.png', device_pixel_ratio=self.devicePixelRatio())
        self.pixmap = self.default_pixmap
        self.pwidth = self.pheight = None
        self.data = {}
        self.last_trim_id = self.last_trim_pixmap = None

        self.do_layout()

    def change_layout(self, vertical):
        self.vertical = vertical
        self.setSizePolicy(QSizePolicy.Policy.Expanding if vertical else QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)

    def value_changed(self, val):
        self.update()

    def setCurrentPixmapSize(self, val):
        self._current_pixmap_size = val

    def minimumSizeHint(self):
        return QSize(100, 100)

    def do_layout(self):
        if self.rect().width() == 0 or self.rect().height() == 0:
            return
        pixmap = self.pixmap
        pwidth, pheight = pixmap.width(), pixmap.height()
        try:
            self.pwidth, self.pheight = fit_image(pwidth, pheight, self.rect().width(), self.rect().height())[1:]
        except Exception:
            self.pwidth, self.pheight = self.rect().width() - 1, self.rect().height() - 1
        self.current_pixmap_size = QSize(self.pwidth, self.pheight)
        self.animation.setEndValue(self.current_pixmap_size)

    def show_data(self, data):
        self.animation.stop()
        same_item = getattr(data, 'id', True) == self.data.get('id', False)
        self.data = {'id': data.get('id', None)}
        if data.cover_data[1]:
            self.pixmap = QPixmap.fromImage(data.cover_data[1])
            if self.pixmap.isNull() or self.pixmap.width() < 5 or self.pixmap.height() < 5:
                self.pixmap = self.default_pixmap
        else:
            self.pixmap = self.default_pixmap
        self.do_layout()
        self.update()
        if not same_item and not config['disable_animations'] and self.isVisible():
            self.animation.start()

    def paintEvent(self, a0):
        canvas_size = self.rect()
        width = self.current_pixmap_size.width()
        extrax = canvas_size.width() - width
        extrax = max(extrax, 0)
        x = int(extrax // 2)
        height = self.current_pixmap_size.height()
        extray = canvas_size.height() - height
        extray = max(extray, 0)
        y = int(extray // 2)
        target = QRect(x, y, width, height)
        p = QPainter(self)
        p.setRenderHints(QPainter.RenderHint.Antialiasing | QPainter.RenderHint.SmoothPixmapTransform)
        try:
            dpr = self.devicePixelRatioF()
        except AttributeError:
            dpr = self.devicePixelRatio()
        spmap = self.pixmap.scaled(target.size() * dpr, Qt.AspectRatioMode.KeepAspectRatio, Qt.TransformationMode.SmoothTransformation)
        spmap.setDevicePixelRatio(dpr)
        with clip_border_radius(p, target):
            p.drawPixmap(target, spmap)
        if gprefs['bd_overlay_cover_size']:
            sztgt = target.adjusted(0, 0, 0, -4)
            f = p.font()
            f.setBold(True)
            p.setFont(f)
            sz = f'\xa0{self.pixmap.width()} x {self.pixmap.height()}\xa0'
            flags = Qt.AlignmentFlag.AlignBottom | Qt.AlignmentFlag.AlignRight | Qt.TextFlag.TextSingleLine
            szrect = p.boundingRect(sztgt, flags, sz)
            p.fillRect(szrect.adjusted(0, 0, 0, 4), QColor(0, 0, 0, 200))
            p.setPen(QPen(QColor(255, 255, 255)))
            p.drawText(sztgt, flags, sz)
        p.end()

    current_pixmap_size = pyqtProperty('QSize', fget=lambda self: self._current_pixmap_size, fset=setCurrentPixmapSize)

    def contextMenuEvent(self, a0):
        cm = QMenu(self)
        paste = cm.addAction(QIcon.ic('edit-paste.png'), _('Paste cover'))
        download = cm.addAction(QIcon.ic('download-metadata.png'), _('Download cover from internet'))
        copy = cm.addAction(QIcon.ic('edit-copy.png'), _('Copy cover'))
        save = cm.addAction(QIcon.ic('save.png'), _('Save cover to disk'))
        remove = cm.addAction(QIcon.ic('trash.png'), _('Remove cover'))
        gc = cm.addAction(QIcon.ic('default_cover.png'), _('Generate cover from metadata'))
        assert paste is not None
        assert download is not None
        assert copy is not None
        assert save is not None
        assert remove is not None
        assert gc is not None
        cm.addSeparator()
        if self.pixmap is not self.default_pixmap and self.data.get('id'):
            book_id = self.data['id']
            tc = QMenu(_('Trim cover'))
            tc.addAction(QIcon.ic('trim.png'), _('Automatically trim borders'), self.trim_cover)
            tc.addAction(_('Trim borders manually'), self.manual_trim_cover)
            tc.addSeparator()
            ac_undo = tc.addAction(QIcon.ic('edit-undo.png'), _('Undo last trim'), self.undo_last_trim)
            assert ac_undo is not None
            ac_undo.setEnabled(self.last_trim_id == book_id)
            cm.addMenu(tc)
            cm.addSeparator()
        _cb = qapplication_or_fail().clipboard()
        assert _cb is not None
        _mime = _cb.mimeData()
        assert _mime is not None
        if not _mime.hasImage():
            paste.setEnabled(False)
        copy.triggered.connect(self.copy_to_clipboard)
        paste.triggered.connect(self.paste_from_clipboard)
        remove.triggered.connect(self.remove_cover)
        download.triggered.connect(self.download_cover)
        gc.triggered.connect(self.generate_cover)
        save.triggered.connect(self.save_cover)
        create_open_cover_with_menu(self, cm)
        m = create_search_internet_menu(self.search_internet.emit)
        cm.addMenu(m)
        cm.exec(a0.globalPos())

    def trim_cover(self):
        book_id = self.data.get('id')
        if not book_id:
            return
        from calibre.utils.img import remove_borders_from_image

        img = image_from_x(self.pixmap)
        nimg = remove_borders_from_image(img)
        if nimg is not img:
            self.last_trim_id = book_id
            self.last_trim_pixmap = self.pixmap
            self.update_cover(QPixmap.fromImage(nimg))

    def manual_trim_cover(self):
        book_id = self.data.get('id')
        if not book_id:
            return
        from calibre.gui2.dialogs.trim_image import TrimImage
        from calibre.utils.img import image_to_data

        cdata = image_to_data(image_from_x(self.pixmap), fmt='PNG', png_compression_level=1)
        d = TrimImage(cdata, parent=self)
        if d.exec() == QDialog.DialogCode.Accepted and d.image_data is not None:
            self.last_trim_id = book_id
            self.last_trim_pixmap = self.pixmap
            self.update_cover(cdata=d.image_data)

    def undo_last_trim(self):
        book_id = self.data.get('id')
        if not book_id or book_id != self.last_trim_id:
            return
        pmap = self.last_trim_pixmap
        self.last_trim_pixmap = self.last_trim_id = None
        self.update_cover(pmap)

    def open_with(self, entry):
        id_ = self.data.get('id', None)
        if id_ is not None:
            self.open_cover_with.emit(id_, entry)

    def choose_open_with(self):
        from calibre.gui2.open_with import choose_program

        entry = choose_program('cover_image', self)
        if entry is not None:
            self.open_with(entry)

    def copy_to_clipboard(self):
        cb = qapplication_or_fail().clipboard()
        assert cb is not None
        cb.setPixmap(self.pixmap)

    def paste_from_clipboard(self, pmap=None):
        if not isinstance(pmap, QPixmap):
            cb = qapplication_or_fail().clipboard()
            assert cb is not None
            pmap = cb.pixmap()
            if pmap.isNull() and cb.supportsSelection():
                pmap = cb.pixmap(QClipboard.Mode.Selection)
            if ismacos:  # Without this there is a crash when Qt tries to save this pixmap as JPEG data
                pmap = pmap.copy()
        if not pmap.isNull():
            self.update_cover(pmap)

    def download_cover(self):
        pmap = download_cover(self, self.data.get('id'), self.pixmap)
        if pmap is not None:
            self.update_cover(pmap)

    def save_cover(self):
        from calibre.gui2.ui import get_gui

        book_id = self.data.get('id')
        db = get_gui(fail_if_absent=True).current_db.new_api
        path = choose_save_file(
            self,
            'save-cover-from-book-details',
            _('Choose cover save location'),
            filters=[(_('JPEG images'), ['jpg', 'jpeg'])],
            all_files=False,
            initial_filename='{}.jpeg'.format(sanitize_file_name(db.field_for('title', book_id, default_value='cover'))),
        )
        if path:
            db.copy_cover_to(book_id, path)

    def update_cover(self, pmap=None, cdata=None):
        if pmap is None:
            pmap = QPixmap()
            pmap.loadFromData(cdata)
        if pmap.isNull():
            return
        if pmap.hasAlphaChannel():
            pmap = QPixmap.fromImage(blend_image(image_from_x(pmap)))
        self.pixmap = pmap
        self.do_layout()
        self.update()
        self.update_tooltip(getattr(self.parent(), 'current_path', ''))
        if not config['disable_animations']:
            self.animation.start()
        id_ = self.data.get('id', None)
        if id_ is not None:
            self.cover_changed.emit(id_, cdata or pixmap_to_data(pmap))

    def generate_cover(self, *args):
        book_id = self.data.get('id')
        if book_id is None:
            return
        from calibre.gui2.ui import get_gui

        mi = get_gui(fail_if_absent=True).current_db.new_api.get_metadata(book_id)
        if not mi.has_cover or confirm(
            _('Are you sure you want to replace the cover? The existing cover will be permanently lost.'),
            'book_details_generate_cover',
        ):
            from calibre.ebooks.covers import generate_cover

            cdata = generate_cover(mi)
            self.update_cover(cdata=cdata)

    def remove_cover(self):
        if not confirm_delete(
            _('Are you sure you want to delete the cover permanently?'),
            'book-details-confirm-cover-remove',
            parent=self,
        ):
            return
        id_ = self.data.get('id', None)
        self.pixmap = self.default_pixmap
        self.do_layout()
        self.update()
        if id_ is not None:
            self.cover_removed.emit(id_)

    def update_tooltip(self, current_path):
        try:
            sz = self.pixmap.size()
        except Exception:
            sz = QSize(0, 0)
        self.setToolTip(
            '<p>'
            + _('Double click to open the Book details window')
            + '<br><br>'
            + _('Path')
            + ': '
            + current_path
            + '<br><br>'
            + _('Cover size: %(width)d x %(height)d pixels') % dict(width=sz.width(), height=sz.height())
        )

    # }}}


# Book Info {{{


class FormatAction(QAction):
    current_fmt: tuple[str | int, str] | None = None
    current_url: str | None = None
    data: tuple[str | None, str | None, str | None] = None, None, None


class BookInfo(HTMLDisplay):
    link_clicked = pyqtSignal(object)
    link_removal_requested = pyqtSignal(object, object)
    remove_format = pyqtSignal(int, object)
    remove_item = pyqtSignal(int, object, object)
    save_format = pyqtSignal(int, object)
    restore_format = pyqtSignal(int, object)
    compare_format = pyqtSignal(int, object)
    set_cover_format = pyqtSignal(int, object)
    copy_link = pyqtSignal(object)
    manage_category = pyqtSignal(object, object)
    open_fmt_with = pyqtSignal(int, object, object)
    edit_book = pyqtSignal(int, object)
    edit_identifiers = pyqtSignal()
    find_in_tag_browser = pyqtSignal(object, object)
    search_internet = pyqtSignal(object)
    search_requested = pyqtSignal(object, object)
    find_in_tag_browser_action: FormatAction
    notes_resource_scheme = RESOURCE_URL_SCHEME

    def __init__(self, vertical, parent=None):
        HTMLDisplay.__init__(self, parent=parent, save_resources_in_document=False)
        self.vertical = vertical
        self.last_rendered_html = '', '', ''
        self.base_url_for_current_book = None
        self.anchor_clicked.connect(self.link_activated)
        for x, icon in [
            ('remove_format', 'trash.png'),
            ('save_format', 'save.png'),
            ('restore_format', 'edit-undo.png'),
            ('copy_link', 'edit-copy.png'),
            ('compare_format', 'diff.png'),
            ('set_cover_format', 'default_cover.png'),
            ('find_in_tag_browser', 'search.png'),
        ]:
            ac = FormatAction(QIcon.ic(icon), '', self)
            ac.triggered.connect(getattr(self, f'{x}_triggerred'))
            setattr(self, f'{x}_action', ac)
        self.manage_action = FormatAction(self)
        self.manage_action.triggered.connect(self.manage_action_triggered)
        self.edit_identifiers_action = QAction(QIcon.ic('identifiers.png'), _('Edit identifiers for this book'), self)
        self.edit_identifiers_action.triggered.connect(self.edit_identifiers)
        self.remove_item_action = ac = FormatAction(QIcon.ic('minus.png'), '...', self)
        ac.data = (None, None, None)
        ac.triggered.connect(self.remove_item_triggered)
        self.copy_identifiers_url_action = ac = FormatAction(QIcon.ic('edit-copy.png'), _('Identifier &URL'), self)
        ac.triggered.connect(self.copy_id_url_triggerred)
        self.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.setDefaultStyleSheet(css())
        if iswindows and not qapplication_or_fail().using_calibre_style:
            # workaround Qt bug that causes selected text to be invisible on
            # windows when using system theme
            bg = self.palette().color(QPalette.ColorRole.Highlight).name()
            fg = self.palette().color(QPalette.ColorRole.HighlightedText).name()
            self.setStyleSheet(f'QTextBrowser {{ selection-color: {fg}; selection-background-color: {bg}; }}')

    def change_layout(self, vertical):
        if vertical != self.vertical:
            self.vertical = vertical
            if self.last_rendered_html[0]:
                html = render_parts(self.last_rendered_html[1], self.last_rendered_html[2], self.vertical)
                self.last_rendered_html = html, self.last_rendered_html[1], self.last_rendered_html[2]
                self.setHtml(html)

    def refresh_css(self):
        self.setDefaultStyleSheet(css(True))

    def remove_item_triggered(self):
        field, value, book_id = self.remove_item_action.data
        if field and confirm(_('Are you sure you want to delete <b>{}</b> from the book?').format(value), 'book_details_remove_item'):
            self.remove_item.emit(book_id, field, value)

    def context_action_triggered(self, which):
        f = getattr(self, f'{which}_action').current_fmt
        url = getattr(self, f'{which}_action').current_url
        if f and 'format' in which:
            book_id, fmt = f
            getattr(self, which).emit(book_id, fmt)
        if url and 'link' in which:
            getattr(self, which).emit(url)

    def remove_format_triggerred(self):
        self.context_action_triggered('remove_format')

    def save_format_triggerred(self):
        self.context_action_triggered('save_format')

    def restore_format_triggerred(self):
        self.context_action_triggered('restore_format')

    def compare_format_triggerred(self):
        self.context_action_triggered('compare_format')

    def set_cover_format_triggerred(self):
        self.context_action_triggered('set_cover_format')

    def copy_link_triggerred(self):
        self.context_action_triggered('copy_link')

    def copy_id_url_triggerred(self):
        if self.copy_identifiers_url_action.current_url:
            self.copy_link.emit(self.copy_identifiers_url_action.current_url)

    def find_in_tag_browser_triggerred(self):
        if self.find_in_tag_browser_action.current_fmt:
            self.find_in_tag_browser.emit(*self.find_in_tag_browser_action.current_fmt)

    def manage_action_triggered(self):
        if self.manage_action.current_fmt:
            self.manage_category.emit(*self.manage_action.current_fmt)

    def link_activated(self, link):
        if str(link.scheme()) in ('http', 'https'):
            return safe_open_url(link)
        link = str(link.toString(NO_URL_FORMATTING))
        self.link_clicked.emit(link)

    def show_data(self, mi):
        html, table, comments = self.last_rendered_html = render_html(mi, self.vertical, self.parent())
        path = getattr(mi, 'path', None)
        self.base_url_for_current_book = QUrl.fromLocalFile(os.path.join(path, 'metadata.opf')) if path else None
        set_html(mi, html, self)

    def get_base_qurl(self):
        return self.base_url_for_current_book

    def process_external_css(self, css):
        return resolve_colors(css)

    def mouseDoubleClickEvent(self, e):
        v = self.viewport()
        assert v is not None
        if v.rect().contains(self.mapFromGlobal(e.globalPos())):
            e.ignore()
        else:
            return HTMLDisplay.mouseDoubleClickEvent(self, e)

    def contextMenuEvent(self, e):
        details_context_menu_event(self, e, self, True)

    def open_with(self, book_id, fmt, entry):
        self.open_fmt_with.emit(book_id, fmt, entry)

    def choose_open_with(self, book_id, fmt):
        from calibre.gui2.open_with import choose_program

        entry = choose_program(fmt, self)
        if entry is not None:
            self.open_with(book_id, fmt, entry)

    def edit_fmt(self, book_id, fmt):
        self.edit_book.emit(book_id, fmt)

    def reindex_fmt(self, book_id, fmt):
        from calibre.gui2.ui import get_gui

        db = get_gui(fail_if_absent=True).current_db.new_api
        if fmt:
            db.reindex_fts_book(book_id, fmt)
        else:
            db.reindex_fts_book(book_id)

    def recount_pages(self, book_id):
        from calibre.gui2.ui import get_gui

        db = get_gui(fail_if_absent=True).current_db.new_api
        for fmt in db.formats(book_id) or ():
            db.format_metadata(book_id, fmt, allow_cache=False, update_db=True)
        db.queue_pages_scan(book_id)
        QTimer.singleShot(100, partial(self.check_for_recount, monotonic(), db.library_id, book_id))

    def check_for_recount(self, start_time: float, library_id: str, book_id: int) -> None:
        from calibre.gui2.ui import get_gui

        db = get_gui(fail_if_absent=True).current_db.new_api
        if db.library_id != library_id or monotonic() - start_time > 10:
            return
        if db.pages_needs_scan((book_id,)):
            QTimer.singleShot(100, partial(self.check_for_recount, start_time, db.library_id, book_id))
            return
        lv = get_gui(fail_if_absent=True).library_view
        current_row = lv.currentIndex().row()
        if lv.current_id != book_id:
            current_row = -1
        lv._model.refresh_ids((book_id,), current_row)


# }}}


class DetailsLayout(QSplitter):  # {{{
    def __init__(self, vertical, parent):
        orientation = Qt.Orientation.Vertical if vertical else Qt.Orientation.Horizontal
        super().__init__(orientation, parent)
        self.vertical = vertical
        self._children = []
        self.setContentsMargins(0, 0, 0, 0)
        self.splitterMoved.connect(self.do_splitter_moved, type=Qt.ConnectionType.QueuedConnection)
        self.resize_timer = QTimer()
        self.resize_timer.setSingleShot(True)
        self.resize_timer.setInterval(5)
        self.resize_timer.timeout.connect(self.do_resize)

    def change_layout(self, vertical):
        self.vertical = vertical
        self.restore_splitter_state()

    def do_resize(self, *args):
        super().resizeEvent(self._resize_ev)
        self.do_layout(self.rect())

    def resizeEvent(self, a0):
        if self.resize_timer.isActive():
            self.resize_timer.stop()
        self._resize_ev = a0
        self.resize_timer.start()

    def minimumSize(self):
        return QSize(190, 200) if self.vertical else QSize(120, 120)

    def addWidget(self, widget):
        if len(self._children) > 2:
            raise ValueError('This layout can only manage two children')
        self._children.append(widget)

    def count(self):
        return len(self._children)

    def sizeHint(self):
        return self.minimumSize()

    @property
    def splitter_state_pref_name(self):
        return 'book_details_widget_splitter_state_' + ('vertical' if self.vertical else 'horizontal')

    def restore_splitter_state(self):
        s = gprefs.get(self.splitter_state_pref_name)
        if s is None:
            s = gprefs.get('book_details_widget_splitter_state')
        if s is None:
            # Without this on first start the splitter is rendered over the cover
            self.setSizes([20, 80])
        else:
            self.restoreState(s)
        self.setOrientation(Qt.Orientation.Vertical if self.vertical else Qt.Orientation.Horizontal)

    def setGeometry(self, *args, **kwargs):
        super().setGeometry(*args, **kwargs)
        self.do_layout(self.geometry())

    def do_splitter_moved(self, *args):
        gprefs[self.splitter_state_pref_name] = bytearray(self.saveState())
        self._children[0].do_layout()

    def cover_height(self, r):
        if not self._children[0].isVisible():
            return 0
        mh = min(int(r.height() // 2), int(4 / 3 * r.width()) + 1)
        try:
            ph = self._children[0].pixmap.height()
        except Exception:
            ph = 0
        if ph > 0:
            mh = min(mh, ph)
        return mh

    def cover_width(self, r):
        if not self._children[0].isVisible():
            return 0
        mw = 1 + int(3 / 4 * r.height())
        try:
            pw = self._children[0].pixmap.width()
        except Exception:
            pw = 0
        if pw > 0:
            mw = min(mw, pw)
        return mw

    def do_layout(self, rect):
        if len(self._children) != 2:
            return
        cm = self.contentsMargins()
        left = cm.left()
        top = cm.top()
        right = cm.right()
        bottom = cm.top()
        r = rect.adjusted(+left, +top, -right, -bottom)
        x = r.x()
        y = r.y()
        cover, details = self._children
        if self.vertical:
            ch = self.cover_height(r)
            cover.setGeometry(QRect(x, y, r.width(), ch))
            y += ch + 5
            details.setGeometry(QRect(x, y, r.width(), r.height() - ch - 5))
        else:
            cw = self.cover_width(r)
            cover.setGeometry(QRect(x, y, cw, r.height()))
            x += cw + 5
            details.setGeometry(QRect(x, y, r.width() - cw - 5, r.height()))
        self.restore_splitter_state()  # only required on first call to do_layout, but ...
        cover.do_layout()


# }}}

# Drag 'n drop {{{


class DropMixin:
    files_dropped = pyqtSignal(object, object)
    remote_file_dropped = pyqtSignal(object, object)

    def __init__(self: DropMixinSelf):
        self.setAcceptDrops(True)

    def dragEnterEvent(self: DropMixinSelf, a0):
        event = a0
        md = event.mimeData()
        if dnd_has_extension(md, image_extensions() + BOOK_EXTENSIONS, allow_all_extensions=True, allow_remote=True) or dnd_has_image(md):
            event.acceptProposedAction()

    def dropEvent(self: DropMixinSelf, a0):
        event = a0
        event.setDropAction(Qt.DropAction.CopyAction)
        md = event.mimeData()

        image_exts = set(image_extensions()) - set(tweaks['cover_drop_exclude'])
        x, y = dnd_get_image(md, image_exts)
        if x is not None:
            # We have an image, set cover
            event.accept()
            if y is None:
                # Local image
                if isinstance(self, DialogBookInfo):
                    self.cover.set_pixmap(x)
                else:
                    self.cover_view.paste_from_clipboard(x)
                    self.update_layout()
            else:
                self.remote_file_dropped.emit(x, y)
                # We do not support setting cover *and* adding formats for
                # a remote drop, anyway, so return
                return

        # Now look for ebook files
        urls, filenames = dnd_get_files(md, BOOK_EXTENSIONS, allow_all_extensions=True, filter_exts=image_exts)
        if not urls:
            # Nothing found
            return

        if not filenames:
            # Local files
            self.files_dropped.emit(event, urls)
        else:
            # Remote files, use the first file
            self.remote_file_dropped.emit(urls[0], filenames[0])
        event.accept()

    def dragMoveEvent(self, a0):
        a0.acceptProposedAction()

    # }}}


class BookDetails(DetailsLayout, DropMixin):  # {{{
    show_book_info = pyqtSignal()
    open_containing_folder = pyqtSignal(int)
    open_data_folder = pyqtSignal(int)
    view_specific_format = pyqtSignal(int, object)
    search_requested = pyqtSignal(object, object)
    remove_specific_format = pyqtSignal(int, object)
    remove_metadata_item = pyqtSignal(int, object, object)
    save_specific_format = pyqtSignal(int, object)
    restore_specific_format = pyqtSignal(int, object)
    set_cover_from_format = pyqtSignal(int, object)
    compare_specific_format = pyqtSignal(int, object)
    copy_link = pyqtSignal(object)
    cover_changed = pyqtSignal(object, object)
    open_cover_with = pyqtSignal(object, object)
    cover_removed = pyqtSignal(object)
    view_device_book = pyqtSignal(object)
    manage_category = pyqtSignal(object, object)
    edit_identifiers = pyqtSignal()
    open_fmt_with = pyqtSignal(int, object, object)
    edit_book = pyqtSignal(int, object)
    find_in_tag_browser = pyqtSignal(object, object)

    def __init__(self, vertical, parent=None):
        super().__init__(vertical, parent)
        DropMixin.__init__(self)
        self.last_data = {}
        self._layout = self
        self.current_path = ''

        self.cover_view = CoverView(vertical, self)
        self.cover_view.search_internet.connect(self.search_internet)
        self.cover_view.cover_changed.connect(self.cover_changed.emit)
        self.cover_view.open_cover_with.connect(self.open_cover_with.emit)
        self.cover_view.cover_removed.connect(self.cover_removed.emit)
        self._layout.addWidget(self.cover_view)
        self.book_info = BookInfo(vertical, self)
        self._layout.addWidget(self.book_info)
        self.book_info.link_clicked.connect(self.handle_click)
        self.book_info.link_removal_requested.connect(self.remove_link)
        self.book_info.remove_format.connect(self.remove_specific_format)
        self.book_info.remove_item.connect(self.remove_metadata_item)
        self.book_info.open_fmt_with.connect(self.open_fmt_with)
        self.book_info.edit_book.connect(self.edit_book)
        self.book_info.save_format.connect(self.save_specific_format)
        self.book_info.restore_format.connect(self.restore_specific_format)
        self.book_info.set_cover_format.connect(self.set_cover_from_format)
        self.book_info.compare_format.connect(self.compare_specific_format)
        self.book_info.copy_link.connect(self.copy_link)
        self.book_info.manage_category.connect(self.manage_category)
        self.book_info.find_in_tag_browser.connect(self.find_in_tag_browser)
        self.book_info.edit_identifiers.connect(self.edit_identifiers)
        self.book_info.search_internet.connect(self.search_internet)
        self.book_info.search_requested.connect(self.search_requested)
        self.setCursor(Qt.CursorShape.PointingHandCursor)

    def change_layout(self, vertical):
        if vertical != self.vertical:
            super().change_layout(vertical)
            self.cover_view.change_layout(vertical)
            self.book_info.change_layout(vertical)
            self.do_layout(self.rect())

    def search_internet(self, data):
        if self.last_data:
            if data.author is None:
                a = self.last_data['authors']
                assert a is not None
                url = url_for_book_search(data.where, title=self.last_data['title'], author=a[0])
            else:
                url = url_for_author_search(data.where, author=data.author)
            safe_open_url(url)

    def handle_click_from_popup(self, link, parent=None):
        parent = parent or self
        typ, val = link.partition(':')[::2]
        from calibre.gui2.ui import get_gui

        db = get_gui(fail_if_absent=True).current_db.new_api

        def search_term(field, val):
            append = ''
            mods = qapplication_or_fail().keyboardModifiers()
            if mods & Qt.KeyboardModifier.ControlModifier:
                append = 'AND' if mods & Qt.KeyboardModifier.ShiftModifier else 'OR'

            fmt = '{}:{}' if is_boolean(field) else '{}:"={}"'
            self.search_requested.emit(fmt.format(field, val.replace('"', '\\"')), append)

        def browse(url):
            if url.startswith(DATA_DIR_NAME + '/') and self.last_data.get('book_id'):
                base = get_gui(fail_if_absent=True).current_db.abspath(self.last_data['book_id'], index_is_id=True)
                path = os.path.join(base, url)
                qurl = QUrl.fromLocalFile(path)
            else:
                qurl = QUrl(url, QUrl.ParsingMode.TolerantMode)
            try:
                safe_open_url(qurl)
            except Exception:
                import traceback

                traceback.print_exc()

        if typ == 'action':
            data = json_loads(from_hex_bytes(val))
            if data.get('url') == NO_SEARCH_LINK:
                return
            dt = data['type']
            if dt == 'search':
                field = data.get('field')
                search_term(data['term'], data['value'])
            elif dt == 'cc_url':
                if data['url']:
                    browse(data['url'])
            elif dt == 'author':
                url = data['url']
                if url == 'calibre':
                    search_term('authors', data['name'])
                else:
                    browse(url)
            elif dt == 'format':
                book_id, fmt = data['book_id'], data['fmt']
                self.view_specific_format.emit(int(book_id), fmt)
            elif dt == 'identifier':
                if data['url']:
                    browse(data['url'])
            elif dt == 'path':
                self.open_containing_folder.emit(int(data['loc']))
            elif dt == 'data-path':
                self.open_data_folder.emit(int(data['loc']))
            elif dt == 'devpath':
                self.view_device_book.emit(data['loc'])
        elif typ == 'notes':
            data = json_loads(from_hex_bytes(val))
            field = data.get('field')
            # It shouldn't be possible for the field to be invalid or the
            # note not to exist, but ...
            if field and db.field_supports_notes(field):
                item_id = data['item_id']
                if item_id is not None and db.notes_for(field, item_id):
                    return self.show_notes(field, item_id, parent)
        else:
            browse(link)

    def handle_click(self, link):
        self.handle_click_from_popup(link)

    def remove_link(self, field, item_value):
        from calibre.gui2.ui import get_gui

        gui = get_gui(fail_if_absent=True)
        db = gui.current_db.new_api
        db.set_link_map(field, {item_value: ''})
        m = gui.library_view._model
        current = gui.library_view.currentIndex()
        m.current_changed(current, current)
        gui.tags_view.recount()

    def show_notes(self, field, item_id, parent=None):
        from calibre.gui2.dialogs.show_category_note import ShowNoteDialog
        from calibre.gui2.ui import get_gui

        ShowNoteDialog(field, item_id, get_gui(fail_if_absent=True).current_db.new_api, parent=parent or self).show()

    def mouseDoubleClickEvent(self, a0):
        a0.accept()
        self.show_book_info.emit()

    def show_data(self, data):
        from calibre.gui2.ui import get_gui

        if getattr(data, 'book_display_info_is_from_device', False) and (gui := get_gui()) and gui.current_view is gui.library_view:
            return
        try:
            self.last_data = {'title': data.title, 'authors': data.authors, 'book_id': getattr(data, 'id', None)}
        except Exception:
            self.last_data = {}
        self.book_info.show_data(data)
        self.cover_view.show_data(data)
        self.current_path = getattr(data, 'path', '')
        self.update_layout()

    def update_layout(self):
        self.cover_view.setVisible(gprefs['bd_show_cover'])
        self._layout.do_layout(self.rect())
        self.cover_view.update_tooltip(self.current_path)

    def reset_info(self):
        self.show_data(Metadata(_('Unknown')))


if TYPE_CHECKING:
    DropMixinSelf = BookDetails | DialogBookInfo

# }}}



#!/usr/bin/env python
# License: GPLv3 Copyright: 2008, Kovid Goyal kovid@kovidgoyal.net

from qt.core import QAbstractItemView, QDialog, QDialogButtonBox, QLabel, QListWidget, QSizePolicy, QVBoxLayout

from calibre.utils.localization import _


class ChoosePluginToolbarsDialog(QDialog):
    def __init__(self, parent, plugin, locations):
        QDialog.__init__(self, parent)
        self.locations = locations

        self.setWindowTitle(_('Add "%s" to toolbars or menus') % plugin.name)

        self._layout = QVBoxLayout(self)
        self.setLayout(self._layout)

        self._header_label = QLabel(_('Select the toolbars and/or menus to add <b>%s</b> to:') % plugin.name)
        self._layout.addWidget(self._header_label)

        self._locations_list = QListWidget(self)
        self._locations_list.setSelectionMode(QAbstractItemView.SelectionMode.MultiSelection)
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Minimum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        self._locations_list.setSizePolicy(sizePolicy)
        for key, text in locations:
            self._locations_list.addItem(text)
            if key in {'toolbar', 'toolbar-device'}:
                _loc_item = self._locations_list.item(self._locations_list.count() - 1)
                assert _loc_item is not None
                _loc_item.setSelected(True)
        self._layout.addWidget(self._locations_list)

        self._footer_label = QLabel(_('You can also customise the plugin locations using <b>Preferences -> Interface -> Toolbars</b>'))
        self._layout.addWidget(self._footer_label)

        button_box = QDialogButtonBox(QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel)
        button_box.accepted.connect(self.accept)
        button_box.rejected.connect(self.reject)
        self._layout.addWidget(button_box)
        self.resize(self.sizeHint())

    def selected_locations(self):
        selected = []
        _sel_model = self._locations_list.selectionModel()
        assert _sel_model is not None
        for row in _sel_model.selectedRows():
            selected.append(self.locations[row.row()])
        return selected



#!/usr/bin/env python
# License: GPLv3 Copyright: 2008, Kovid Goyal kovid@kovidgoyal.net

"""
Miscelleaneous utilities.
"""

from time import time

from polyglot.builtins import as_bytes


def join_with_timeout(q, timeout=2):
    """Join the queue q with a specified timeout. Blocks until all tasks on
    the queue are done or times out with a runtime error."""
    q.all_tasks_done.acquire()
    try:
        endtime = time() + timeout
        while q.unfinished_tasks:
            remaining = endtime - time()
            if remaining <= 0.0:
                raise RuntimeError('Waiting for queue to clear timed out')
            q.all_tasks_done.wait(remaining)
    finally:
        q.all_tasks_done.release()


def unpickle_binary_string(data):
    # Maintains compatibility with python's pickle module protocol version 2
    import struct

    PROTO, SHORT_BINSTRING, BINSTRING = b'\x80', b'U', b'T'
    if data.startswith(PROTO + b'\x02'):
        offset = 2
        which = data[offset : offset + 1]
        offset += 1
        if which == BINSTRING:
            (sz,) = struct.unpack_from('<i', data, offset)
            offset += struct.calcsize('<i')
        elif which == SHORT_BINSTRING:
            sz = ord(data[offset : offset + 1])
            offset += 1
        else:
            return
        return data[offset : offset + sz]


def pickle_binary_string(data):
    # Maintains compatibility with python's pickle module protocol version 2
    import struct

    PROTO, STOP, BINSTRING = b'\x80', b'.', b'T'
    data = as_bytes(data)
    return PROTO + b'\x02' + BINSTRING + struct.pack(b'<i', len(data)) + data + STOP


def kill_parent_if_needed(parent_process_handle: int, timeout: float = 1.0, exit_code: int = 0) -> None:
    import ctypes

    WAIT_TIMEOUT = 0x00000102

    kernel32 = ctypes.windll.kernel32
    try:
        if kernel32.WaitForSingleObject(parent_process_handle, int(timeout * 1000)) == WAIT_TIMEOUT:
            kernel32.TerminateProcess(parent_process_handle, exit_code)
    finally:
        kernel32.CloseHandle(parent_process_handle)



# License: GPLv3 Copyright: 2009, John Schember <john@nachtimwald.com>

import os

from calibre.customize.conversion import OptionRecommendation, OutputFormatPlugin
from calibre.utils.localization import _


class TCROutput(OutputFormatPlugin):
    name = 'TCR Output'
    author = 'John Schember'
    file_type = 'tcr'
    commit_name = 'tcr_output'

    options = {
        OptionRecommendation(
            name='tcr_output_encoding',
            recommended_value='utf-8',
            level=OptionRecommendation.LOW,
            help=_('Specify the character encoding of the output document. The default is utf-8.'),
        )
    }

    def convert(self, oeb_book, output, input_plugin, opts, log):
        output_path = output
        from calibre.ebooks.compression.tcr import compress
        from calibre.ebooks.txt.txtml import TXTMLizer

        close = False
        if not hasattr(output_path, 'write'):
            close = True
            if not os.path.exists(os.path.dirname(output_path)) and os.path.dirname(output_path):
                os.makedirs(os.path.dirname(output_path))
            out_stream = open(output_path, 'wb')
        else:
            out_stream = output_path

        setattr(opts, 'flush_paras', False)
        setattr(opts, 'max_line_length', 0)
        setattr(opts, 'force_max_line_length', False)
        setattr(opts, 'indent_paras', False)

        writer = TXTMLizer(log)
        opts.use_alt_text_for_images = False
        txt = writer.extract_content(oeb_book, opts).encode(opts.tcr_output_encoding, 'replace')

        log.info('Compressing text...')
        txt = compress(txt)

        out_stream.seek(0)
        out_stream.truncate()
        out_stream.write(txt)

        if close:
            out_stream.close()



#!/usr/bin/env python
# License: GPLv3 Copyright: 2008, Kovid Goyal <kovid at kovidgoyal.net>

from qt.core import QDialog, QLineEdit, Qt

from calibre.gui2 import error_dialog
from calibre.gui2.dialogs.smartdevice_ui import Ui_Dialog
from calibre.utils.localization import _
from calibre.utils.mdns import get_all_ips


def ipaddr_sort_key(ipaddr):
    if '.' in ipaddr:
        parts = tuple(map(int, ipaddr.split('.')))
        is_private = parts[0] in (192, 170, 10)
        return (0 if is_private else 1), parts


def get_all_ip_addresses():
    ipaddrs = []
    for iface in get_all_ips().values():
        for addrs in iface:
            if 'broadcast' in addrs and addrs['addr'] != '127.0.0.1':
                ipaddrs.append(addrs['addr'])
    ipaddrs.sort(key=ipaddr_sort_key)
    return ipaddrs


class SmartdeviceDialog(QDialog, Ui_Dialog):
    def __init__(self, parent):
        QDialog.__init__(self, parent)
        Ui_Dialog.__init__(self)
        self.setupUi(self)

        self.password_box.setToolTip(
            '<p>'
            + _(
                'Use a password if calibre is running on a network that '
                'is not secure. For example, if you run calibre on a laptop, '
                'use that laptop in an airport, and want to connect your '
                'smart device to calibre, you should use a password.'
            )
            + '</p>'
        )

        self.autostart_box.setToolTip(
            '<p>'
            + _(
                'Check this box if you want calibre to automatically start the '
                'smart device interface when calibre starts. You should not do '
                'this if you are using a network that is not secure and you '
                'are not setting a password.'
            )
            + '</p>'
        )

        self.use_fixed_port.setToolTip(
            '<p>'
            + _(
                'Check this box if you want calibre to use a fixed network '
                'port. Normally you will not need to do this. However, if '
                'your device consistently fails to connect to calibre, '
                'try checking this box and entering a number.'
            )
            + '</p>'
        )

        self.fixed_port.setToolTip(
            '<p>'
            + _('Try 9090. If calibre says that it fails to connect to the port, try another number. You can use any number between 8,000 and 65,535.')
            + '</p>'
        )

        self.ip_addresses.setToolTip(
            '<p>'
            + _(
                'These are the IP addresses for this computer. If you decide to have your device connect to '
                'calibre using a fixed IP address, one of these addresses should '
                'be the one you use. It is unlikely but possible that the correct '
                'IP address is not listed here, in which case you will need to go '
                "to your computer's control panel to get a complete list of "
                "your computer's network interfaces and IP addresses."
            )
            + '</p>'
        )

        self.show_password.stateChanged[int].connect(self.toggle_password)
        self.use_fixed_port.stateChanged[int].connect(self.use_fixed_port_changed)

        self.device_manager = parent.device_manager

        if self.device_manager.get_option('smartdevice', 'autostart'):
            self.autostart_box.setChecked(True)

        pw = self.device_manager.get_option('smartdevice', 'password')
        if pw:
            self.password_box.setText(pw)

        self.orig_fixed_port = self.device_manager.get_option('smartdevice', 'use_fixed_port')
        self.orig_port_number = self.device_manager.get_option('smartdevice', 'port_number')
        self.fixed_port.setText(self.orig_port_number)
        self.use_fixed_port.setChecked(self.orig_fixed_port)
        if not self.orig_fixed_port:
            self.fixed_port.setEnabled(False)

        if pw:
            self.password_box.setText(pw)

        forced_ip = self.device_manager.get_option('smartdevice', 'force_ip_address')
        if forced_ip:
            self.ip_addresses.setText(forced_ip)
        else:
            self.ip_addresses.setText(', '.join(get_all_ip_addresses()))

        self.resize(self.sizeHint())

    def use_fixed_port_changed(self, state):
        self.fixed_port.setEnabled(Qt.CheckState(state) == Qt.CheckState.Checked)

    def toggle_password(self, state):
        self.password_box.setEchoMode(QLineEdit.EchoMode.Password if state == Qt.CheckState.Unchecked else QLineEdit.EchoMode.Normal)

    def accept(self):
        port = str(self.fixed_port.text())
        if not port:
            error_dialog(self, _('Invalid port number'), _('You must provide a port number.'), show=True)
            return
        try:
            port = int(port)
        except Exception:
            error_dialog(self, _('Invalid port number'), _('The port must be a number between 8000 and 65535.'), show=True)
            return

        if port < 8000 or port > 65535:
            error_dialog(self, _('Invalid port number'), _('The port must be a number between 8000 and 65535.'), show=True)
            return

        self.device_manager.set_option('smartdevice', 'password', str(self.password_box.text()))
        self.device_manager.set_option('smartdevice', 'autostart', self.autostart_box.isChecked())
        self.device_manager.set_option('smartdevice', 'use_fixed_port', self.use_fixed_port.isChecked())
        self.device_manager.set_option('smartdevice', 'port_number', str(self.fixed_port.text()))

        message = self.device_manager.start_plugin('smartdevice')

        if not self.device_manager.is_running('smartdevice'):
            error_dialog(
                self,
                _('Problem starting the wireless device'),
                _('The wireless device driver had problems starting. It said "%s"') % message,
                show=True,
            )
            self.device_manager.set_option('smartdevice', 'use_fixed_port', self.orig_fixed_port)
            self.device_manager.set_option('smartdevice', 'port_number', self.orig_port_number)
        else:
            QDialog.accept(self)



# autogenerated by __main__.py do not edit
top_level_module_names = (
    'QtCore',
    'QtGui',
    'QtWidgets',
    'QtNetwork',
    'QtSvg',
    'QtPrintSupport',
    'QtOpenGL',
    'QtOpenGLWidgets',
    'QtQuick',
    'QtMultimedia',
    'QtMultimediaWidgets',
    'QtTextToSpeech',
    'QtWebEngineCore',
    'QtWebEngineWidgets',
    'QtWebChannel',
    'QtDBus',
)


def __getattr__(name):
    if name in top_level_module_names:
        import importlib

        return importlib.import_module('PyQt6.' + name)
    raise AttributeError(name)



# License: GPLv3 Copyright: 2011, John Schember <john@nachtimwald.com>

from functools import partial

from qt.core import QMenu, QSize, Qt, QTreeView

from calibre.customize.ui import store_plugins
from calibre.gui2.metadata.single_download import RichTextDelegate
from calibre.gui2.store.config.chooser.models import Delegate, Matches
from calibre.utils.localization import _


class ResultsView(QTreeView):
    def __init__(self, parent=None):
        QTreeView.__init__(self, parent)

        self._model = Matches(list(store_plugins()))
        self.setModel(self._model)

        self.setIconSize(QSize(24, 24))

        self.rt_delegate = RichTextDelegate(self)
        self.delegate = Delegate()
        self.setItemDelegate(self.delegate)

        for i in self._model.HTML_COLS:
            self.setItemDelegateForColumn(i, self.rt_delegate)

        for i in range(self._model.columnCount()):
            self.resizeColumnToContents(i)

        self._model.sort(1, Qt.SortOrder.AscendingOrder)
        _header = self.header()
        assert _header is not None
        _header.setSortIndicator(self._model.sort_col, self._model.sort_order)

    def contextMenuEvent(self, a0):
        index = self.indexAt(a0.pos())

        if not index.isValid():
            return

        _m = self.model()
        assert isinstance(_m, Matches)
        plugin = _m.get_plugin(index)

        menu = QMenu(self)
        ca = menu.addAction(_('Configure...'), partial(self.configure_plugin, plugin))
        if not plugin.is_customizable():
            assert ca is not None
            ca.setEnabled(False)
        menu.exec(a0.globalPos())

    def configure_plugin(self, plugin):
        plugin.do_user_config(self)
